'''
Views related to information about a given function.
'''
from PyQt5 import QtCore
from PyQt5.QtWidgets import QMenu, QWidget

import vivisect.qt.ctxmenu as viv_q_ctxmenu
import visgraph.renderers.qgraphtree as vg_qgraphtree

from vqt.basics import *
from vqt.application import VQDockWidget


class FuncBlockModel(BasicModel):
    columns = ('Base Address', 'Size')


class FunctionBlocksView(BasicTreeView):

    _sig_BlockSelected = QtCore.pyqtSignal(object)

    def __init__(self, vw, parent=None):
        self.vw = vw

        BasicTreeView.__init__(self, parent=parent)
        self.setModel(FuncBlockModel())
        self.setWindowTitle('Code Blocks: ')

    def selectionChanged(self, selected, unselected):

        indexes = selected.indexes()
        if indexes:
            index = indexes[0]
            row = self.model().rows[index.row()]

            block = row[-1]
            self._sig_BlockSelected.emit(block)

            cmap = {}
            bva, bsize, fva = block
            for i in range(bsize):
                cmap[bva + i] = 'yellow'

            # Since we have a reference to the GUI, lets also
            # send a coloration signal.
            vwgui = self.vw.getVivGui()
            vwgui.vivMemColorSignal.emit(cmap)

        return BasicTreeView.selectionChanged(self, selected, unselected)

    def closeEvent(self, event):
        # On close, remove any color mappings...
        vwgui = self.vw.getVivGui()
        vwgui.vivMemColorSignal.emit({})
        return BasicTreeView.closeEvent(self, event)

    def functionSelected(self, fva):
        self.setWindowTitle('Code Blocks: %s' % self.vw.getName(fva))
        blocks = self.vw.getFunctionBlocks(fva)
        rows = [('0x%.8x' % block[0], block[1], block) for block in blocks]
        model = FuncBlockModel(rows=rows)
        self.setModel(model)
        self.sortByColumn(0, QtCore.Qt.AscendingOrder)


class FuncCallsView(QWidget):

    def __init__(self, vw, parent=None):
        self.vw = vw
        super().__init__(parent=parent)

        self.graphview = vg_qgraphtree.QGraphTreeView(None, (), parent=self)
        self.graphview._sig_NodeContextMenu.connect(self.nodeContextMenu)
        self.setLayout(VBox(self.graphview))

    def functionSelected(self, fva):
        self.setWindowTitle('Call Graph: %s' % self.vw.getName(fva))
        nprops = self.vw._call_graph.getNodeProps(fva)
        self.graphview.loadNewGraph(self.vw._call_graph, ((fva, nprops), ))

    def nodeContextMenu(self, pos, nid, nprops):
        menu = QMenu(parent=self)
        viv_q_ctxmenu.buildContextMenu(self.vw, va=nid, menu=menu)
        menu.exec_(pos)



"""
FreeBSD support...
"""

import os
import ctypes
import ctypes.util as cutil
import logging

import envi.cli as e_cli
import envi.const as e_const

import vtrace
import vtrace.archs.i386 as v_i386
import vtrace.archs.amd64 as v_amd64
import vtrace.platforms.base as v_base
import vtrace.platforms.posix as v_posix
import vtrace.util as v_util

logger = logging.getLogger(__name__)

libc = ctypes.CDLL(cutil.find_library("c"))
libkvm = ctypes.CDLL(cutil.find_library("kvm"))

# kvm_getprocs cmds
KERN_PROC_ALL           = 0       # everything
KERN_PROC_PID           = 1       # by process id
KERN_PROC_PGRP          = 2       # by process group id
KERN_PROC_SESSION       = 3       # by session of pid
KERN_PROC_TTY           = 4       # by controlling tty
KERN_PROC_UID           = 5       # by effective uid
KERN_PROC_RUID          = 6       # by real uid
KERN_PROC_ARGS          = 7       # get/set arguments/proctitle
KERN_PROC_PROC          = 8       # only return procs
KERN_PROC_SV_NAME       = 9       # get syscall vector name
KERN_PROC_RGID          = 10      # by real group id
KERN_PROC_GID           = 11      # by effective group id
KERN_PROC_PATHNAME      = 12      # path to executable
KERN_PROC_INC_THREAD    = 0x10    # Include threads in filtered results

pid_t = ctypes.c_int32
lwpid_t = ctypes.c_int32
void_p = ctypes.c_void_p
dev_t = ctypes.c_uint32
sigset_t = ctypes.c_uint32*4
uid_t = ctypes.c_uint32
gid_t = ctypes.c_uint32
fixpt_t = ctypes.c_uint32
caddr_t = ctypes.c_void_p
vm_size_t = ctypes.c_ulong
segsz_t = ctypes.c_ulong

# Could go crazy and grep headers for this stuff ;)
KI_NGROUPS = 16
OCOMMLEN = 16
WMESGLEN = 8
LOGNAMELEN = 17
LOCKNAMELEN = 8
COMMLEN = 19
KI_EMULNAMELEN = 16
KI_NSPARE_INT = 10
KI_NSPARE_PTR = 7
KI_NSPARE_LONG = 12


def c_buf(size):
    return ctypes.c_char * size

class PRIORITY(ctypes.Structure):
    _fields_ = (
        ("pri_class", ctypes.c_ubyte),
        ("pri_level", ctypes.c_ubyte),
        ("pri_native", ctypes.c_ubyte),
        ("pri_user", ctypes.c_ubyte)
    )

class TIMEVAL(ctypes.Structure):
    _fields_ = (
        ("tv_sec", ctypes.c_long),
        ("tv_usec", ctypes.c_long)
    )

class RUSAGE(ctypes.Structure):
    _fields_ = (
        ("ru_utime", TIMEVAL),          # user time used
        ("ru_stime", TIMEVAL),          # system time used
        ("ru_maxrss", ctypes.c_long),   #
        ("ru_ixrss", ctypes.c_long),    # (j) integral shared memory size
        ("ru_idrss", ctypes.c_long),    # (j) integral unshared data
        ("ru_isrss", ctypes.c_long),    # (j) integral unshared stack
        ("ru_minflt", ctypes.c_long),   # (c) page reclaims
        ("ru_majflt", ctypes.c_long),   # (c) page faults
        ("ru_nswap", ctypes.c_long),    # (c + j) swaps
        ("ru_inblock", ctypes.c_long),  # (n) block input operations
        ("ru_oublock", ctypes.c_long),  # (n) block output operations
        ("ru_msgsnd", ctypes.c_long),   # (n) messages sent
        ("ru_msgrcv", ctypes.c_long),   # (n) messages received
        ("ru_nsignals", ctypes.c_long), # (c) signals received
        ("ru_nvcsw", ctypes.c_long),    # (j) voluntary context switches
        ("ru_nivcsw", ctypes.c_long),   # (j) involuntary
    )


class KINFO_PROC(ctypes.Structure):
    _fields_ = (
        ("ki_structsize", ctypes.c_int),# size of this structure
        ("ki_layout", ctypes.c_int),    # reserved: layout identifier
        ("ki_args", void_p),            # address of command arguments (struct pargs*)
        ("ki_paddr", void_p),           # address of proc (struct proc*)
        ("ki_addr", void_p),            # kernel virtual addr of u-area (struct user*)
        ("ki_tracep", void_p),          # pointer to trace file (struct vnode *)
        ("ki_textvp", void_p),          # pointer to executable file (struct vnode *)
        ("ki_fd", void_p),              # pointer to open file info (struct filedesc  *)
        ("ki_vmspace", void_p),         # pointer to kernel vmspace struct (struct vmspace *)
        ("ki_wchan", void_p),           # sleep address (void*)
        ("ki_pid", pid_t),              # Process identifier
        ("ki_ppid", pid_t),             # parent process id
        ("ki_pgid", pid_t),             # process group id
        ("ki_tpgid", pid_t),            # tty process group id
        ("ki_sid", pid_t),              # Process session ID
        ("ki_tsid", pid_t),             # Terminal session ID
        ("ki_jobc", ctypes.c_short),    # job control counter
        ("ki_spare_short1", ctypes.c_short), #
        ("ki_tdev", dev_t),             # controlling tty dev
        ("ki_siglist", sigset_t),       # Signals arrived but not delivered
        ("ki_sigmask", sigset_t),       # Current signal mask
        ("ki_sigignore", sigset_t),     # Signals being ignored
        ("ki_sigcatch", sigset_t),      # Signals being caught by user
        ("ki_uid", uid_t),              # effective user id
        ("ki_ruid", uid_t),             # Real user id
        ("ki_svuid", uid_t),            # Saved effective user id
        ("ki_rgid", gid_t),             # Real group id
        ("ki_svgid", gid_t),            # Saved effective group id
        ("ki_ngroups", ctypes.c_short), # number of groups
        ("ki_spare_short2", ctypes.c_short),
        ("ki_groups", gid_t * KI_NGROUPS), # groups
        ("ki_size", vm_size_t),         # virtual size
        ("ki_rssize", segsz_t),         # current resident set size in pages
        ("ki_swrss", segsz_t),          # resident set size before last swap
        ("ki_tsize", segsz_t),          # text size (pages) XXX
        ("ki_dsize", segsz_t),          # data size (pages) XXX
        ("ki_ssize", segsz_t),          # stack size (pages)
        ("ki_xstat", ctypes.c_ushort),  # Exit status for wait and stop signal
        ("ki_acflag", ctypes.c_ushort), # Accounting flags
        ("ki_pctcpu", fixpt_t),         # %cpu for process during ki_swtime
        ("ki_estcpu", ctypes.c_uint),   # Time averaged value of ki_cpticks
        ("ki_slptime", ctypes.c_uint),  # Time since last blocked
        ("ki_swtime", ctypes.c_uint),   # Time swapped in or out
        ("ki_spareint1", ctypes.c_int), # unused (just here for alignment)
        ("ki_runtime", ctypes.c_uint64),# Real time in microsec
        ("ki_start", TIMEVAL),          # starting time
        ("ki_childtime", TIMEVAL),      # time used by process children
        ("ki_flag", ctypes.c_long),     # P_* flags
        ("ki_kiflag", ctypes.c_long),   # KI_* flags
        ("ki_traceflag", ctypes.c_int), # kernel trace points
        ("ki_stat", ctypes.c_char),     # S* process status
        ("ki_nice", ctypes.c_ubyte),    # Process "nice" value
        ("ki_lock", ctypes.c_char),     # Process lock (prevent swap) count
        ("ki_rqindex", ctypes.c_char),  # Run queue index
        ("ki_oncpu", ctypes.c_char),    # Which cpu we are on
        ("ki_lastcpu", ctypes.c_char),  # Last cpu we were on
        ("ki_ocomm", c_buf(OCOMMLEN+1)),      # command name
        ("ki_wmesg", c_buf(WMESGLEN+1)),      # wchan message
        ("ki_login", c_buf(LOGNAMELEN+1)),    # setlogin name
        ("ki_lockname", c_buf(LOCKNAMELEN+1)),# lock name
        ("ki_comm", c_buf(COMMLEN+1)),        # command name
        ("ki_emul", c_buf(KI_EMULNAMELEN+1)), # emulation name
        ("ki_sparestrings",c_buf(68)),   # spare string space
        ("ki_spareints", ctypes.c_int*KI_NSPARE_INT),
        ("ki_jid", ctypes.c_int),       # Process jail ID
        ("ki_numthreads", ctypes.c_int),# KSE number of total threads
        ("ki_tid", lwpid_t),            # thread id
        ("ki_pri", PRIORITY),           # process priority
        ("ki_rusage", RUSAGE),          # process rusage statistics
        # XXX - most fields in ki_rusage_ch are not (yet) filled in
        ("ki_rusage_ch", RUSAGE),       # rusage of children processes
        ("ki_pcb", void_p),             # kernel virtual addr of pcb
        ("ki_kstack", void_p),          # kernel virtual addr of stack
        ("ki_udata", void_p),           # User convenience pointer
        ("ki_spareptrs", void_p*KI_NSPARE_PTR),
        ("ki_sparelongs", ctypes.c_long*KI_NSPARE_LONG),
        ("ki_sflag", ctypes.c_long),    # PS_* flags
        ("ki_tdflags", ctypes.c_long),  # KSE kthread flag
    )

libkvm.kvm_getprocs.argtypes = [caddr_t, ctypes.c_int, ctypes.c_int, caddr_t]
libkvm.kvm_getprocs.restype = ctypes.POINTER(KINFO_PROC)

libkvm.kvm_open.argtypes = [ctypes.c_char_p, ctypes.c_char_p, ctypes.c_char_p, ctypes.c_int, ctypes.c_char_p]
libkvm.kvm_open.restype = caddr_t

# All the FreeBSD ptrace defines
PT_TRACE_ME     = 0       #/* child declares it's being traced */
PT_READ_I       = 1       #/* read word in child's I space */
PT_READ_D       = 2       #/* read word in child's D space */
PT_WRITE_I      = 4       #/* write word in child's I space */
PT_WRITE_D      = 5       #/* write word in child's D space */
PT_CONTINUE     = 7       #/* continue the child */
PT_KILL         = 8       #/* kill the child process */
PT_STEP         = 9       #/* single step the child */
PT_ATTACH       = 10      #/* trace some running process */
PT_DETACH       = 11      #/* stop tracing a process */
PT_IO           = 12      #/* do I/O to/from stopped process. */
PT_LWPINFO      = 13      #/* Info about the LWP that stopped. */
PT_GETNUMLWPS   = 14      #/* get total number of threads */
PT_GETLWPLIST   = 15      #/* get thread list */
PT_CLEARSTEP    = 16      #/* turn off single step */
PT_SETSTEP      = 17      #/* turn on single step */
PT_SUSPEND      = 18      #/* suspend a thread */
PT_RESUME       = 19      #/* resume a thread */
PT_TO_SCE       = 20      # Stop on syscall entry
PT_TO_SCX       = 21      # Stop on syscall exit
PT_SYSCALL      = 22      # Stop on syscall entry and exit
PT_GETREGS      = 33      #/* get general-purpose registers */
PT_SETREGS      = 34      #/* set general-purpose registers */
PT_GETFPREGS    = 35      #/* get floating-point registers */
PT_SETFPREGS    = 36      #/* set floating-point registers */
PT_GETDBREGS    = 37      #/* get debugging registers */
PT_SETDBREGS    = 38      #/* set debugging registers */
#PT_FIRSTMACH    = 64      #/* for machine-specific requests */

# On PT_IO addr is a pointer to a struct

class PTRACE_IO_DESC(ctypes.Structure):
    _fields_ = [
        ("piod_op", ctypes.c_int),      # I/O operation
        ("piod_offs", ctypes.c_void_p), # Child offset
        ("piod_addr", ctypes.c_void_p), # Parent Offset
        ("piod_len", ctypes.c_uint)     # Size
    ]

# Operations in piod_op.
PIOD_READ_D     = 1       # Read from D space
PIOD_WRITE_D    = 2       # Write to D space
PIOD_READ_I     = 3       # Read from I space
PIOD_WRITE_I    = 4       # Write to I space

class PTRACE_LWPINFO(ctypes.Structure):
    _fields_ = (
        ("pl_lwpid", lwpid_t),
        ("pl_event", ctypes.c_int),
        ("pl_flags", ctypes.c_int),
        ("pl_sigmask", sigset_t),
        ("pl_siglist", sigset_t),
    )

PL_EVENT_NONE   = 0
PL_EVENT_SIGNAL = 1

PL_FLAGS_SA    = 0
PL_FLAGS_BOUND = 1

class FreeBSDMixin:

    def __init__(self):
        self.initMode("Syscall", False, "Break on Syscalls")
        self.kvmh = libkvm.kvm_open(None, None, None, 0, "vtrace")
        if not os.path.exists('/proc/curproc/file'):
            raise Exception("VDB needs /proc! (use: mount -t procfs procfs /proc)")

    def finiMixin(self):
        if self.kvmh is not None:
            libkvm.kvm_close(self.kvmh)

    def platformReadMemory(self, address, size):
        #FIXME optimize for speed!
        iod = PTRACE_IO_DESC()
        buf = ctypes.create_string_buffer(size)

        iod.piod_op = PIOD_READ_D
        iod.piod_addr = ctypes.addressof(buf)
        iod.piod_offs = address
        iod.piod_len = size

        if v_posix.ptrace(PT_IO, self.pid, ctypes.addressof(iod), 0) != 0:
            raise Exception("ptrace PT_IO failed to read 0x%.8x" % address)

        return buf.raw

    def platformWriteMemory(self, address, buf):
        #FIXME optimize for speed!
        iod = PTRACE_IO_DESC()

        cbuf = ctypes.create_string_buffer(buf)

        iod.piod_op = PIOD_WRITE_D
        iod.piod_addr = ctypes.addressof(cbuf)
        iod.piod_offs = address
        iod.piod_len = len(buf)

        if v_posix.ptrace(PT_IO, self.pid, ctypes.addressof(iod), 0) != 0:
            raise Exception("ptrace PT_IO failed to write 0x%.8x" % address)

    @v_base.threadwrap
    def platformAttach(self, pid):
        if v_posix.ptrace(PT_ATTACH, pid, 0, 0) != 0:
            raise Exception("Ptrace Attach Failed")

    def _getExeName(self, pid):
        return os.readlink('/proc/%d/file' % pid)

    #@v_base.threadwrap
    def platformExec(self, cmdline):
        # Basically just like the one in the Ptrace mixin...
        self.execing = True
        cmdlist = e_cli.splitargs(cmdline)
        os.stat(cmdlist[0])
        pid = os.fork()
        if pid == 0:
            v_posix.ptrace(PT_TRACE_ME, 0, 0, 0)
            os.execv(cmdlist[0], cmdlist)
            sys.exit(-1)
        return pid

    def handleAttach(self):
        self.setMeta('ExeName', self._getExeName(self.pid))
        return v_posix.PosixMixin.handleAttach(self)

    @v_base.threadwrap
    def platformWait(self):

        pid,status = v_posix.PosixMixin.platformWait(self)

        # Get the thread id from the ptrace interface
        info = PTRACE_LWPINFO()
        size = ctypes.sizeof(info)
        if v_posix.ptrace(PT_LWPINFO, self.pid, ctypes.addressof(info), size) == 0:
            self.setMeta('ThreadId', info.pl_lwpid)

        return pid,status

    @v_base.threadwrap
    def platformProcessEvent(self, event):

        pid,status = event

        if os.WIFEXITED(status):

            exitcode = os.WEXITSTATUS(status)

            tid = self.getMeta("ThreadId", None)
            if tid is None or len(self.getThreads()) == 0:
                self._fireExit( exitcode )
                return

            self._fireExitThread(tid, exitcode)

            # set thread to pid ( the thread exited... so... )
            self.setMeta('ThreadId', pid)
            self._fireExit( exitcode )

        elif os.WIFSIGNALED(status):
            self._fireExit( os.WTERMSIG( status ) )

        elif os.WIFSTOPPED(status):
            sig = os.WSTOPSIG(status)
            self.handlePosixSignal(sig)

        else:
            logger.warning('Received unknown event: %s', event)


    @v_base.threadwrap
    def platformStepi(self):
        self.stepping = True
        if v_posix.ptrace(PT_STEP, self.pid, 1, 0) != 0:
            raise Exception("ptrace PT_STEP failed!")

    @v_base.threadwrap
    def platformContinue(self):
        cmd = PT_CONTINUE
        if self.getMode("Syscall"):
            cmd = PT_SYSCALL

        sig = self.getCurrentSignal()
        if sig is None:
            sig = 0

        # In freebsd address is the place to continue from
        # but 1 means use existing EIP
        if v_posix.ptrace(cmd, self.pid, 1, sig) != 0:
            raise Exception("ptrace PT_CONTINUE/PT_SYSCALL failed")

    @v_base.threadwrap
    def platformDetach(self):
        if v_posix.ptrace(PT_DETACH, self.pid, 1, 0) != 0:
            raise Exception("Ptrace Detach Failed")

    @v_base.threadwrap
    def platformGetThreads(self):
        ret = {}
        cnt = self._getThreadCount()
        if cnt & 0xffffffff == 0xffffffff:
            return ret
        buf = (ctypes.c_int * cnt)()
        if v_posix.ptrace(PT_GETLWPLIST, self.pid, ctypes.addressof(buf), cnt) != cnt:
            raise Exception("ptrace PW_GETLWPLIST failed")
        for x in buf:
            ret[x] = x
        return ret

    @v_base.threadwrap
    def platformSuspendThread(self, tid):
        if v_posix.ptrace(PT_SUSPEND, tid, 0, 0) != 0:
            raise Exception("ptrace PT_SUSPEND failed!")

    @v_base.threadwrap
    def platformResumeThread(self, tid):
        if v_posix.ptrace(PT_RESUME, tid, 0, 0) != 0:
            raise Exception("ptrace PT_RESUME failed!")

    def _getThreadCount(self):
        return v_posix.ptrace(PT_GETNUMLWPS, self.pid, 0, 0)

    def platformGetFds(self):
        return []

    def platformGetMaps(self):
        # FIXME make this not need proc
        ret = []
        mpath = "/proc/%d/map" % self.pid

        with open(mpath, 'rb') as mapfile:
            for line in mapfile:
                perms = 0
                fname = ""
                maptup = line.split(None)
                base = int(maptup[0], 16)
                max  = int(maptup[1], 16)
                permstr = maptup[5]

                if maptup[11] == "vnode":
                    fname = maptup[12].strip()

                if permstr[0] == 'r':
                    perms |= e_const.MM_READ

                if permstr[1] == 'w':
                    perms |= e_const.MM_WRITE

                if permstr[2] == 'x':
                    perms |= e_const.MM_EXEC

                ret.append((base, max-base, perms, fname))

        return ret

    def platformPs(self):
        ret = []
        cnt = ctypes.c_uint(0)

        p = libkvm.kvm_getprocs(self.kvmh, KERN_PROC_PROC, 0, ctypes.addressof(cnt))
        for i in range(cnt.value):
            kinfo = p[i]
            if kinfo.ki_structsize != ctypes.sizeof(KINFO_PROC):
                logger.warning("WARNING: KINFO_PROC CHANGED SIZE, Trying to account for it... good luck")
            ret.append((kinfo.ki_pid, kinfo.ki_comm))

        ret.reverse()
        return ret

class bsd_regs_i386(ctypes.Structure):
    _fields_ = (
        ("fs",  ctypes.c_ulong),
        ("es",  ctypes.c_ulong),
        ("ds",  ctypes.c_ulong),
        ("edi",  ctypes.c_ulong),
        ("esi",  ctypes.c_ulong),
        ("ebp",  ctypes.c_ulong),
        ("isp",  ctypes.c_ulong),
        ("ebx",  ctypes.c_ulong),
        ("edx",  ctypes.c_ulong),
        ("ecx",  ctypes.c_ulong),
        ("eax",  ctypes.c_ulong),
        ("trapno",  ctypes.c_ulong),
        ("err",  ctypes.c_ulong),
        ("eip",  ctypes.c_ulong),
        ("cs",  ctypes.c_ulong),
        ("eflags",  ctypes.c_ulong),
        ("esp",  ctypes.c_ulong),
        ("ss",  ctypes.c_ulong),
        ("gs",  ctypes.c_ulong),
        ("debug0",  ctypes.c_ulong),
        ("debug1",  ctypes.c_ulong),
        ("debug2",  ctypes.c_ulong),
        ("debug3",  ctypes.c_ulong),
        ("debug4",  ctypes.c_ulong),
        ("debug5",  ctypes.c_ulong),
        ("debug6",  ctypes.c_ulong),
        ("debug7",  ctypes.c_ulong),
    )

i386_DBG_OFF = (19*4)

class bsd_regs_amd64(ctypes.Structure):
    _fields_ = (
        ("r15", ctypes.c_ulonglong),
        ("r14", ctypes.c_ulonglong),
        ("r13", ctypes.c_ulonglong),
        ("r12", ctypes.c_ulonglong),
        ("r11", ctypes.c_ulonglong),
        ("r10", ctypes.c_ulonglong),
        ("r9", ctypes.c_ulonglong),
        ("r8", ctypes.c_ulonglong),
        ("rdi", ctypes.c_ulonglong),
        ("rsi", ctypes.c_ulonglong),
        ("rbp", ctypes.c_ulonglong),
        ("rbx", ctypes.c_ulonglong),
        ("rdx", ctypes.c_ulonglong),
        ("rcx", ctypes.c_ulonglong),
        ("rax", ctypes.c_ulonglong),
        ("trapno", ctypes.c_ulonglong),
        ("err", ctypes.c_ulonglong),
        ("rip", ctypes.c_ulonglong),
        ("cs", ctypes.c_ulonglong),
        ("rflags", ctypes.c_ulonglong),
        ("rsp", ctypes.c_ulonglong),
        ("ss", ctypes.c_ulonglong),
        ("debug0", ctypes.c_ulonglong),
        ("debug1", ctypes.c_ulonglong),
        ("debug2", ctypes.c_ulonglong),
        ("debug3", ctypes.c_ulonglong),
        ("debug4", ctypes.c_ulonglong),
        ("debug5", ctypes.c_ulonglong),
        ("debug6", ctypes.c_ulonglong),
        ("debug7", ctypes.c_ulonglong),
        ("debug8", ctypes.c_ulonglong),
        ("debug9", ctypes.c_ulonglong),
        ("debug10", ctypes.c_ulonglong),
        ("debug11", ctypes.c_ulonglong),
        ("debug12", ctypes.c_ulonglong),
        ("debug13", ctypes.c_ulonglong),
        ("debug14", ctypes.c_ulonglong),
        ("debug15", ctypes.c_ulonglong),
    )

amd64_DBG_OFF = (22*ctypes.sizeof(ctypes.c_uint64))

class FreeBSDi386Trace(
    vtrace.Trace,
    FreeBSDMixin,
    v_i386.i386Mixin,
    v_posix.ElfMixin,
    v_posix.PosixMixin,
    v_base.TracerBase):

    def __init__(self):
        vtrace.Trace.__init__(self)
        v_base.TracerBase.__init__(self)
        v_posix.ElfMixin.__init__(self)
        v_posix.PosixMixin.__init__(self)
        v_i386.i386Mixin.__init__(self)
        FreeBSDMixin.__init__(self)

    @v_base.threadwrap
    def platformGetRegCtx(self, tid):
        ctx = self.archGetRegCtx()
        u = bsd_regs_i386()

        addr = ctypes.addressof(u)
        if v_posix.ptrace(PT_GETREGS, tid, addr, 0) != 0:
            raise Exception("ptrace PT_GETREGS failed!")

        if v_posix.ptrace(PT_GETDBREGS, tid, addr+i386_DBG_OFF, 0) != 0:
            raise Exception("ptrace PT_GETDBREGS failed!")

        ctx._rctx_Import(u)

        return ctx

    @v_base.threadwrap
    def platformSetRegCtx(self, tid, ctx):
        u = bsd_regs_i386()
        ctx._rctx_Export(u)
        addr = ctypes.addressof(u)
        if v_posix.ptrace(PT_SETREGS, self.pid, addr, 0) != 0:
            raise Exception("ptrace PT_SETREGS failed!")
        if v_posix.ptrace(PT_SETDBREGS, self.pid, addr+i386_DBG_OFF, 0) != 0:
            raise Exception("ptrace PT_SETDBREGS failed!")


class FreeBSDAmd64Trace(
    vtrace.Trace,
    FreeBSDMixin,
    v_amd64.Amd64Mixin,
    v_posix.ElfMixin,
    v_posix.PosixMixin,
    v_base.TracerBase):

    def __init__(self):
        vtrace.Trace.__init__(self)
        v_base.TracerBase.__init__(self)
        v_posix.ElfMixin.__init__(self)
        v_posix.PosixMixin.__init__(self)
        v_amd64.Amd64Mixin.__init__(self)
        FreeBSDMixin.__init__(self)

    def _getAmdRegsStruct(self, tid):
        '''
        Get (and populate) a register structure
        (even set regs needs to get it first...)
        '''
        u = bsd_regs_amd64()
        addr = ctypes.addressof(u)
        if v_posix.ptrace(PT_GETREGS, tid, addr, 0) != 0:
            raise Exception("ptrace PT_GETREGS failed!")
        if v_posix.ptrace(PT_GETDBREGS, tid, addr+amd64_DBG_OFF, 0) != 0:
            raise Exception("ptrace PT_GETDBREGS failed!")
        return u

    @v_base.threadwrap
    def platformGetRegCtx(self, tid):
        ctx = self.archGetRegCtx()
        u = self._getAmdRegsStruct(tid)
        ctx._rctx_Import(u)
        return ctx

    @v_base.threadwrap
    def platformSetRegCtx(self, tid, ctx):
        u = self._getAmdRegsStruct(tid)
        ctx._rctx_Export(u)
        addr = ctypes.addressof(u)
        if v_posix.ptrace(PT_SETREGS, tid, addr, 0) != 0:
            raise Exception("ptrace PT_SETREGS failed!")
        if v_posix.ptrace(PT_SETDBREGS, tid, addr+amd64_DBG_OFF, 0) != 0:
            raise Exception("ptrace PT_SETDBREGS failed!")





'''
A function analysis module that determines if any of the leaf nodes in a function
have a clean return

Depends on the codeblock and thunk passes being finished, so typically this will be the last
pass to run.

Generally, idea is to search through the terminal nodes of a function, and if they end in a
noreturn call, or just straight up don't have a return
* And if all the terminal nodes end in that, then there's no way the function returns, and we should
bail
'''
import logging

import envi
import vivisect.const as v_const
import vivisect.tools.graphutil as v_t_graph


logger = logging.getLogger(__name__)


def analyzeFunction(vw, fva):
    if vw.isFunctionThunk(fva):
        # Don't bother with import thunks
        for ref in vw.getXrefsFrom(fva, rtype=v_const.REF_CODE):
            loc = vw.getLocation(ref[v_const.XR_TO])
            if loc and loc[v_const.L_LTYPE] == v_const.LOC_IMPORT:
                return

    try:
        g = v_t_graph.buildFunctionGraph(vw, fva)
    except Exception as exc:
        logger.warning('NoRet: Failed to build function graph for function 0x%x: %s', fva, exc)
        return
    _, _, leaves = v_t_graph.getNodeWeightHisto(g)

    hasret = False
    for weight, nodes in leaves.items():
        for node in nodes:
            cbstart, cbsize, cbfunc = vw.getCodeBlock(node[0])
            cbend = cbstart + cbsize - 1

            lva, lsize, ltype, linfo = vw.getLocation(cbend)
            if ltype != v_const.LOC_OP:
                pass
            if vw.isNoReturnVa(lva):
                continue
            if linfo & envi.IF_RET:
                hasret = True
                break
            # be wary of dynamic branches we couldn't resolve
            if linfo & envi.IF_BRANCH:
                hasret = True
                break

    # 0x14006ba90 out of omnetpp.exe at O2, 64bit is a good counter example (that shouldn't be no ret)
    if not hasret:
        logger.info('Marking 0x%.8x as no return' % fva)
        vw.addNoReturnVa(fva)



# APIs for Windows 32-bit rpcrt4 library
# Format:  rettype, retname, callconv, exactname, arglist(type, name)
#          arglist type is one of ['int', 'void *']
#          arglist name is one of [None, 'funcptr', 'obj', 'ptr']

api_defs = {
    'rpcrt4.main_entry':( 'int', None, 'stdcall', 'rpcrt4.main_entry', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.cstdstubbuffer_addref':( 'int', None, 'stdcall', 'rpcrt4.CStdStubBuffer_AddRef', (('int', None),) ),
    'rpcrt4.cstdstubbuffer_connect':( 'int', None, 'stdcall', 'rpcrt4.CStdStubBuffer_Connect', (('int', None), ('int', None)) ),
    'rpcrt4.cstdstubbuffer_countrefs':( 'int', None, 'stdcall', 'rpcrt4.CStdStubBuffer_CountRefs', (('int', None),) ),
    'rpcrt4.cstdstubbuffer_debugserverqueryinterface':( 'int', None, 'stdcall', 'rpcrt4.CStdStubBuffer_DebugServerQueryInterface', (('int', None), ('int', None)) ),
    'rpcrt4.cstdstubbuffer_debugserverrelease':( 'int', None, 'stdcall', 'rpcrt4.CStdStubBuffer_DebugServerRelease', (('int', None), ('int', None)) ),
    'rpcrt4.cstdstubbuffer_disconnect':( 'int', None, 'stdcall', 'rpcrt4.CStdStubBuffer_Disconnect', (('int', None),) ),
    'rpcrt4.cstdstubbuffer_invoke':( 'int', None, 'stdcall', 'rpcrt4.CStdStubBuffer_Invoke', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.cstdstubbuffer_isiidsupported':( 'int', None, 'stdcall', 'rpcrt4.CStdStubBuffer_IsIIDSupported', (('int', None), ('int', None)) ),
    'rpcrt4.cstdstubbuffer_queryinterface':( 'int', None, 'stdcall', 'rpcrt4.CStdStubBuffer_QueryInterface', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.checkverificationtrailer':( 'int', None, 'stdcall', 'rpcrt4.CheckVerificationTrailer', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.createproxyfromtypeinfo':( 'int', None, 'stdcall', 'rpcrt4.CreateProxyFromTypeInfo', (('int', None), ('int', None), ('void *', 'ptr'), ('int', None), ('int', None)) ),
    'rpcrt4.createstubfromtypeinfo':( 'int', None, 'stdcall', 'rpcrt4.CreateStubFromTypeInfo', (('int', None), ('void *', 'ptr'), ('int', None), ('int', None)) ),
    'rpcrt4.dceerrorinqtexta':( 'int', None, 'stdcall', 'rpcrt4.DceErrorInqTextA', (('int', None), ('int', None)) ),
    'rpcrt4.dceerrorinqtextw':( 'int', None, 'stdcall', 'rpcrt4.DceErrorInqTextW', (('int', None), ('int', None)) ),
    'rpcrt4.dllgetclassobject':( 'int', None, 'stdcall', 'rpcrt4.DllGetClassObject', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.dllregisterserver':( 'int', None, 'cdecl', 'rpcrt4.DllRegisterServer', () ),
    'rpcrt4.globalmutexclearexternal':( 'int', None, 'cdecl', 'rpcrt4.GlobalMutexClearExternal', () ),
    'rpcrt4.globalmutexrequestexternal':( 'int', None, 'cdecl', 'rpcrt4.GlobalMutexRequestExternal', () ),
    'rpcrt4.iunknown_addref_proxy':( 'int', None, 'stdcall', 'rpcrt4.IUnknown_AddRef_Proxy', (('int', None),) ),
    'rpcrt4.iunknown_queryinterface_proxy':( 'int', None, 'stdcall', 'rpcrt4.IUnknown_QueryInterface_Proxy', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.iunknown_release_proxy':( 'int', None, 'stdcall', 'rpcrt4.IUnknown_Release_Proxy', (('int', None),) ),
    'rpcrt4.i_rpcallocate':( 'int', None, 'stdcall', 'rpcrt4.I_RpcAllocate', (('DWORD', None),) ),
    'rpcrt4.i_rpcasyncabortcall':( 'int', None, 'stdcall', 'rpcrt4.I_RpcAsyncAbortCall', (('int', None), ('int', None)) ),
    'rpcrt4.i_rpcasyncsethandle':( 'int', None, 'stdcall', 'rpcrt4.I_RpcAsyncSetHandle', (('int', None), ('int', None)) ),
    'rpcrt4.i_rpcbcacheallocate':( 'int', None, 'stdcall', 'rpcrt4.I_RpcBCacheAllocate', (('DWORD', None),) ),
    'rpcrt4.i_rpcbcachefree':( 'int', None, 'stdcall', 'rpcrt4.I_RpcBCacheFree', (('void *', 'ptr'),) ),
    'rpcrt4.i_rpcbindingcopy':( 'int', None, 'stdcall', 'rpcrt4.I_RpcBindingCopy', ( ('void *', 'ptr'),) ),
    'rpcrt4.i_rpcbindinghandletoasynchandle':( 'int', None, 'stdcall', 'rpcrt4.I_RpcBindingHandleToAsyncHandle', ( ('int', None),) ),
    'rpcrt4.i_rpcbindinginqconnid':( 'int', None, 'stdcall', 'rpcrt4.I_RpcBindingInqConnId', ( ('int', None), ('int', None)) ),
    'rpcrt4.i_rpcbindinginqdynamicendpointa':( 'int', None, 'stdcall', 'rpcrt4.I_RpcBindingInqDynamicEndpointA', (('int', None), ('int', None)) ),
    'rpcrt4.i_rpcbindinginqdynamicendpointw':( 'int', None, 'stdcall', 'rpcrt4.I_RpcBindingInqDynamicEndpointW', ( ('void *', 'ptr'),) ),
    'rpcrt4.i_rpcbindinginqlocalclientpid':( 'int', None, 'stdcall', 'rpcrt4.I_RpcBindingInqLocalClientPID', (('int', None), ('int', None)) ),
    'rpcrt4.i_rpcbindinginqsecuritycontext':( 'int', None, 'stdcall', 'rpcrt4.I_RpcBindingInqSecurityContext', ( ('int', None),) ),
    'rpcrt4.i_rpcbindinginqtransporttype':( 'int', None, 'stdcall', 'rpcrt4.I_RpcBindingInqTransportType', (('int', None), ('int', None)) ),
    'rpcrt4.i_rpcbindinginqwireidforsnego':( 'int', None, 'stdcall', 'rpcrt4.I_RpcBindingInqWireIdForSnego', ( ('int', None),) ),
    'rpcrt4.i_rpcbindingisclientlocal':( 'int', None, 'stdcall', 'rpcrt4.I_RpcBindingIsClientLocal', ( ('int', None),) ),
    'rpcrt4.i_rpcbindingtostaticstringbindingw':( 'int', None, 'stdcall', 'rpcrt4.I_RpcBindingToStaticStringBindingW', ( ('void *', 'ptr'),) ),
    'rpcrt4.i_rpcclearmutex':( 'int', None, 'cdecl', 'rpcrt4.I_RpcClearMutex', () ),
    'rpcrt4.i_rpcconnectioninqsockbuffsize':( 'int', None, 'stdcall', 'rpcrt4.I_RpcConnectionInqSockBuffSize', (('int', None), ('int', None)) ),
    'rpcrt4.i_rpcconnectionsetsockbuffsize':( 'int', None, 'stdcall', 'rpcrt4.I_RpcConnectionSetSockBuffSize', (('int', None), ('int', None)) ),
    'rpcrt4.i_rpcdeletemutex':( 'int', None, 'stdcall', 'rpcrt4.I_RpcDeleteMutex', () ),
    'rpcrt4.i_rpcenablewmitrace':( 'int', None, 'stdcall', 'rpcrt4.I_RpcEnableWmiTrace', (('int', None), ('int', None)) ),
    'rpcrt4.i_rpcexceptionfilter':( 'int', None, 'stdcall', 'rpcrt4.I_RpcExceptionFilter', (('int', None),) ),
    'rpcrt4.i_rpcfree':( 'int', None, 'stdcall', 'rpcrt4.I_RpcFree', (('void *', 'ptr'),) ),
    'rpcrt4.i_rpcfreebuffer':( 'int', None, 'stdcall', 'rpcrt4.I_RpcFreeBuffer', (('int', None),) ),
    'rpcrt4.i_rpcfreepipebuffer':( 'int', None, 'stdcall', 'rpcrt4.I_RpcFreePipeBuffer', (('int', None),) ),
    'rpcrt4.i_rpcgetbuffer':( 'int', None, 'stdcall', 'rpcrt4.I_RpcGetBuffer', (('int', None),) ),
    'rpcrt4.i_rpcgetbufferwithobject':( 'int', None, 'stdcall', 'rpcrt4.I_RpcGetBufferWithObject', (('int', None), ('void *', 'obj')) ),
    'rpcrt4.i_rpcgetcurrentcallhandle':( 'int', None, 'cdecl', 'rpcrt4.I_RpcGetCurrentCallHandle', () ),
    'rpcrt4.i_rpcgetextendederror':( 'int', None, 'cdecl', 'rpcrt4.I_RpcGetExtendedError', () ),
    'rpcrt4.i_rpcifinqtransfersyntaxes':( 'int', None, 'stdcall', 'rpcrt4.I_RpcIfInqTransferSyntaxes', (('int', None), ('void *', 'ptr'), ('int', None), ('void *', 'ptr')) ),
    'rpcrt4.i_rpclogevent':( 'int', None, 'stdcall', 'rpcrt4.I_RpcLogEvent', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.i_rpcmapwin32status':( 'int', None, 'stdcall', 'rpcrt4.I_RpcMapWin32Status', (('int', None),) ),
    'rpcrt4.i_rpcnegotiatetransfersyntax':( 'int', None, 'stdcall', 'rpcrt4.I_RpcNegotiateTransferSyntax', (('int', None),) ),
    'rpcrt4.i_rpcnsbindingsetentrynamea':( 'int', None, 'stdcall', 'rpcrt4.I_RpcNsBindingSetEntryNameA', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.i_rpcnsbindingsetentrynamew':( 'int', None, 'stdcall', 'rpcrt4.I_RpcNsBindingSetEntryNameW', ( ('int', None), ('int', None)) ),
    'rpcrt4.i_rpcnsinterfaceexported':( 'int', None, 'stdcall', 'rpcrt4.I_RpcNsInterfaceExported', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.i_rpcnsinterfaceunexported':( 'int', None, 'stdcall', 'rpcrt4.I_RpcNsInterfaceUnexported', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.i_rpcparsesecurity':( 'int', None, 'stdcall', 'rpcrt4.I_RpcParseSecurity', (('int', None), ('void *', 'ptr')) ),
    'rpcrt4.i_rpcpauseexecution':( 'int', None, 'cdecl', 'rpcrt4.I_RpcPauseExecution', () ),
    'rpcrt4.i_rpcproxynewconnection':( 'int', None, 'stdcall', 'rpcrt4.I_RpcProxyNewConnection', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.i_rpcreallocpipebuffer':( 'int', None, 'stdcall', 'rpcrt4.I_RpcReallocPipeBuffer', (('int', None), ('int', None)) ),
    'rpcrt4.i_rpcreceive':( 'int', None, 'stdcall', 'rpcrt4.I_RpcReceive', (('int', None), ('int', None)) ),
    'rpcrt4.i_rpcrequestmutex':( 'int', None, 'stdcall', 'rpcrt4.I_RpcRequestMutex', (('int', None),) ),
    'rpcrt4.i_rpcsend':( 'int', None, 'stdcall', 'rpcrt4.I_RpcSend', (('int', None),) ),
    'rpcrt4.i_rpcsendreceive':( 'int', None, 'stdcall', 'rpcrt4.I_RpcSendReceive', (('int', None),) ),
    'rpcrt4.i_rpcserverallocateipport':( 'int', None, 'stdcall', 'rpcrt4.I_RpcServerAllocateIpPort', (('int', None), ('int', None)) ),
    'rpcrt4.i_rpcservercheckclientrestriction':( 'int', None, 'stdcall', 'rpcrt4.I_RpcServerCheckClientRestriction', () ),
    'rpcrt4.i_rpcserverinqaddresschangefn':( 'int', None, 'cdecl', 'rpcrt4.I_RpcServerInqAddressChangeFn', () ),
    'rpcrt4.i_rpcserverinqlocalconnaddress':( 'int', None, 'stdcall', 'rpcrt4.I_RpcServerInqLocalConnAddress', ( ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.i_rpcserverinqtransporttype':( 'int', None, 'stdcall', 'rpcrt4.I_RpcServerInqTransportType', (('int', None),) ),
    'rpcrt4.i_rpcserverregisterforwardfunction':( 'int', None, 'stdcall', 'rpcrt4.I_RpcServerRegisterForwardFunction', (('int', None),) ),
    'rpcrt4.i_rpcserversetaddresschangefn':( 'int', None, 'stdcall', 'rpcrt4.I_RpcServerSetAddressChangeFn', (('int', None),) ),
    'rpcrt4.i_rpcserveruseprotseq2a':( 'int', None, 'stdcall', 'rpcrt4.I_RpcServerUseProtseq2A', (('int', None), ('int', None), ('int', None), ('void *', 'ptr'), ('void *', 'ptr')) ),
    'rpcrt4.i_rpcserveruseprotseq2w':( 'int', None, 'stdcall', 'rpcrt4.I_RpcServerUseProtseq2W', (('int', None), ('void *', 'ptr'), ('int', None), ('void *', 'ptr'), ('int', None)) ),
    'rpcrt4.i_rpcserveruseprotseqep2a':( 'int', None, 'stdcall', 'rpcrt4.I_RpcServerUseProtseqEp2A', (('int', None), ('int', None), ('int', None), ('int', None), ('void *', 'ptr'), ('void *', 'ptr')) ),
    'rpcrt4.i_rpcserveruseprotseqep2w':( 'int', None, 'stdcall', 'rpcrt4.I_RpcServerUseProtseqEp2W', (('int', None), ('void *', 'ptr'), ('int', None), ('void *', 'ptr'), ('void *', 'ptr'), ('int', None)) ),
    'rpcrt4.i_rpcsessionstrictcontexthandle':( 'int', None, 'cdecl', 'rpcrt4.I_RpcSessionStrictContextHandle', () ),
    'rpcrt4.i_rpcssdontserializecontext':( 'int', None, 'cdecl', 'rpcrt4.I_RpcSsDontSerializeContext', () ),
    'rpcrt4.i_rpcsystemfunction001':( 'int', None, 'stdcall', 'rpcrt4.I_RpcSystemFunction001', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.i_rpctransconnectionallocatepacket':( 'int', None, 'stdcall', 'rpcrt4.I_RpcTransConnectionAllocatePacket', (('int', None), ('DWORD', None)) ),
    'rpcrt4.i_rpctransconnectionfreepacket':( 'int', None, 'stdcall', 'rpcrt4.I_RpcTransConnectionFreePacket', (('int', None), ('void *', 'ptr')) ),
    'rpcrt4.i_rpctransconnectionreallocpacket':( 'int', None, 'stdcall', 'rpcrt4.I_RpcTransConnectionReallocPacket', (('int', None), ('int', None), ('int', None), ('DWORD', None)) ),
    'rpcrt4.i_rpctransdatagramallocate':( 'int', None, 'stdcall', 'rpcrt4.I_RpcTransDatagramAllocate', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.i_rpctransdatagramallocate2':( 'int', None, 'stdcall', 'rpcrt4.I_RpcTransDatagramAllocate2', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.i_rpctransdatagramfree':( 'int', None, 'stdcall', 'rpcrt4.I_RpcTransDatagramFree', (('int', None), ('void *', 'ptr')) ),
    'rpcrt4.i_rpctransgetthreadevent':( 'int', None, 'cdecl', 'rpcrt4.I_RpcTransGetThreadEvent', () ),
    'rpcrt4.i_rpctransiocancelled':( 'int', None, 'stdcall', 'rpcrt4.I_RpcTransIoCancelled', ( ('void *', 'ptr'),) ),
    'rpcrt4.i_rpctransservernewconnection':( 'int', None, 'stdcall', 'rpcrt4.I_RpcTransServerNewConnection', () ),
    'rpcrt4.i_rpcturnoneeinfopropagation':( 'int', None, 'cdecl', 'rpcrt4.I_RpcTurnOnEEInfoPropagation', () ),
    'rpcrt4.i_uuidcreate':( 'int', None, 'stdcall', 'rpcrt4.I_UuidCreate', (('int', None),) ),
    'rpcrt4.midl_wchar_strcpy':( 'int', None, 'stdcall', 'rpcrt4.MIDL_wchar_strcpy', (('int', None), ('int', None)) ),
    'rpcrt4.midl_wchar_strlen':( 'int', None, 'stdcall', 'rpcrt4.MIDL_wchar_strlen', (('int', None),) ),
    'rpcrt4.mesbufferhandlereset':( 'int', None, 'stdcall', 'rpcrt4.MesBufferHandleReset', (('int', None), ('int', None), ('int', None), ('void *', 'ptr'), ('int', None), ('void *', 'ptr')) ),
    'rpcrt4.mesdecodebufferhandlecreate':( 'int', None, 'stdcall', 'rpcrt4.MesDecodeBufferHandleCreate', (('int', None), ('int', None), ('void *', 'ptr')) ),
    'rpcrt4.mesdecodeincrementalhandlecreate':( 'int', None, 'stdcall', 'rpcrt4.MesDecodeIncrementalHandleCreate', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.mesencodedynbufferhandlecreate':( 'int', None, 'stdcall', 'rpcrt4.MesEncodeDynBufferHandleCreate', (('void *', 'ptr'), ('void *', 'ptr'), ('void *', 'ptr')) ),
    'rpcrt4.mesencodefixedbufferhandlecreate':( 'int', None, 'stdcall', 'rpcrt4.MesEncodeFixedBufferHandleCreate', (('int', None), ('int', None), ('void *', 'ptr'), ('void *', 'ptr')) ),
    'rpcrt4.mesencodeincrementalhandlecreate':( 'int', None, 'stdcall', 'rpcrt4.MesEncodeIncrementalHandleCreate', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.meshandlefree':( 'int', None, 'stdcall', 'rpcrt4.MesHandleFree', (('void *', 'ptr'),) ),
    'rpcrt4.mesincrementalhandlereset':( 'int', None, 'stdcall', 'rpcrt4.MesIncrementalHandleReset', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.mesinqprocencodingid':( 'int', None, 'stdcall', 'rpcrt4.MesInqProcEncodingId', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrccontextbinding':( 'int', None, 'stdcall', 'rpcrt4.NDRCContextBinding', (('int', None),) ),
    'rpcrt4.ndrccontextmarshall':( 'int', None, 'stdcall', 'rpcrt4.NDRCContextMarshall', (('int', None), ('int', None)) ),
    'rpcrt4.ndrccontextunmarshall':( 'int', None, 'stdcall', 'rpcrt4.NDRCContextUnmarshall', (('int', None),) ),
    'rpcrt4.ndrscontextmarshall':( 'int', None, 'stdcall', 'rpcrt4.NDRSContextMarshall', ( ('int', None), ('int', None)) ),
    'rpcrt4.ndrscontextmarshall2':( 'int', None, 'stdcall', 'rpcrt4.NDRSContextMarshall2', ( ('void *', 'obj'), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrscontextmarshallex':( 'int', None, 'stdcall', 'rpcrt4.NDRSContextMarshallEx', ( ('void *', 'obj'), ('int', None), ('int', None)) ),
    'rpcrt4.ndrscontextunmarshall':( 'int', None, 'stdcall', 'rpcrt4.NDRSContextUnmarshall', (('int', None), ('int', None)) ),
    'rpcrt4.ndrscontextunmarshall2':( 'int', None, 'stdcall', 'rpcrt4.NDRSContextUnmarshall2', ( ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrscontextunmarshallex':( 'int', None, 'stdcall', 'rpcrt4.NDRSContextUnmarshallEx', ( ('int', None), ('int', None)) ),
    'rpcrt4.ndrcopy':( 'int', None, 'stdcall', 'rpcrt4.NDRcopy', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrallocate':( 'int', None, 'stdcall', 'rpcrt4.NdrAllocate', (('int', None), ('int', None)) ),
    'rpcrt4.ndrasyncclientcall':( 'int', None, 'cdecl', 'rpcrt4.NdrAsyncClientCall', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrasyncservercall':( 'int', None, 'stdcall', 'rpcrt4.NdrAsyncServerCall', (('int', None),) ),
    'rpcrt4.ndrbytecountpointerbuffersize':( 'int', None, 'stdcall', 'rpcrt4.NdrByteCountPointerBufferSize', (('void *', 'ptr'), ('int', None), ('int', None)) ),
    'rpcrt4.ndrbytecountpointerfree':( 'int', None, 'stdcall', 'rpcrt4.NdrByteCountPointerFree', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrbytecountpointermarshall':( 'int', None, 'stdcall', 'rpcrt4.NdrByteCountPointerMarshall', (('void *', 'ptr'), ('int', None), ('int', None)) ),
    'rpcrt4.ndrbytecountpointerunmarshall':( 'int', None, 'stdcall', 'rpcrt4.NdrByteCountPointerUnmarshall', (('void *', 'ptr'), ('void *', 'ptr'), ('int', None), ('int', None)) ),
    'rpcrt4.ndrcstdstubbuffer2_release':( 'int', None, 'stdcall', 'rpcrt4.NdrCStdStubBuffer2_Release', (('int', None), ('int', None)) ),
    'rpcrt4.ndrcstdstubbuffer_release':( 'int', None, 'stdcall', 'rpcrt4.NdrCStdStubBuffer_Release', (('int', None), ('int', None)) ),
    'rpcrt4.ndrclearoutparameters':( 'int', None, 'stdcall', 'rpcrt4.NdrClearOutParameters', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrclientcall':( 'int', None, 'cdecl', 'rpcrt4.NdrClientCall', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrclientcall2':( 'int', None, 'cdecl', 'rpcrt4.NdrClientCall2', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrclientcontextmarshall':( 'int', None, 'stdcall', 'rpcrt4.NdrClientContextMarshall', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrclientcontextunmarshall':( 'int', None, 'stdcall', 'rpcrt4.NdrClientContextUnmarshall', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrclientinitialize':( 'int', None, 'stdcall', 'rpcrt4.NdrClientInitialize', (('void *', 'ptr'), ('void *', 'ptr'), ('int', None), ('int', None)) ),
    'rpcrt4.ndrclientinitializenew':( 'int', None, 'stdcall', 'rpcrt4.NdrClientInitializeNew', (('void *', 'ptr'), ('void *', 'ptr'), ('int', None), ('int', None)) ),
    'rpcrt4.ndrcomplexarraybuffersize':( 'int', None, 'stdcall', 'rpcrt4.NdrComplexArrayBufferSize', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrcomplexarrayfree':( 'int', None, 'stdcall', 'rpcrt4.NdrComplexArrayFree', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrcomplexarraymarshall':( 'int', None, 'stdcall', 'rpcrt4.NdrComplexArrayMarshall', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrcomplexarraymemorysize':( 'int', None, 'stdcall', 'rpcrt4.NdrComplexArrayMemorySize', (('int', None), ('void *', 'obj')) ),
    'rpcrt4.ndrcomplexarrayunmarshall':( 'int', None, 'stdcall', 'rpcrt4.NdrComplexArrayUnmarshall', (('void *', 'ptr'), ('void *', 'ptr'), ('void *', 'obj'), ('int', None)) ),
    'rpcrt4.ndrcomplexstructbuffersize':( 'int', None, 'stdcall', 'rpcrt4.NdrComplexStructBufferSize', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrcomplexstructfree':( 'int', None, 'stdcall', 'rpcrt4.NdrComplexStructFree', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrcomplexstructmarshall':( 'int', None, 'stdcall', 'rpcrt4.NdrComplexStructMarshall', (('void *', 'ptr'), ('int', None), ('int', None)) ),
    'rpcrt4.ndrcomplexstructmemorysize':( 'int', None, 'stdcall', 'rpcrt4.NdrComplexStructMemorySize', (('int', None), ('void *', 'obj')) ),
    'rpcrt4.ndrcomplexstructunmarshall':( 'int', None, 'stdcall', 'rpcrt4.NdrComplexStructUnmarshall', (('void *', 'ptr'), ('void *', 'ptr'), ('void *', 'obj'), ('int', None)) ),
    'rpcrt4.ndrconformantarraybuffersize':( 'int', None, 'stdcall', 'rpcrt4.NdrConformantArrayBufferSize', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrconformantarrayfree':( 'int', None, 'stdcall', 'rpcrt4.NdrConformantArrayFree', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrconformantarraymarshall':( 'int', None, 'stdcall', 'rpcrt4.NdrConformantArrayMarshall', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrconformantarraymemorysize':( 'int', None, 'stdcall', 'rpcrt4.NdrConformantArrayMemorySize', (('int', None), ('int', None)) ),
    'rpcrt4.ndrconformantarrayunmarshall':( 'int', None, 'stdcall', 'rpcrt4.NdrConformantArrayUnmarshall', (('void *', 'ptr'), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrconformantstringbuffersize':( 'int', None, 'stdcall', 'rpcrt4.NdrConformantStringBufferSize', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrconformantstringmarshall':( 'int', None, 'stdcall', 'rpcrt4.NdrConformantStringMarshall', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrconformantstringmemorysize':( 'int', None, 'stdcall', 'rpcrt4.NdrConformantStringMemorySize', (('int', None), ('int', None)) ),
    'rpcrt4.ndrconformantstringunmarshall':( 'int', None, 'stdcall', 'rpcrt4.NdrConformantStringUnmarshall', (('int', None), ('void *', 'ptr'), ('int', None), ('int', None)) ),
    'rpcrt4.ndrconformantstructbuffersize':( 'int', None, 'stdcall', 'rpcrt4.NdrConformantStructBufferSize', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrconformantstructfree':( 'int', None, 'stdcall', 'rpcrt4.NdrConformantStructFree', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrconformantstructmarshall':( 'int', None, 'stdcall', 'rpcrt4.NdrConformantStructMarshall', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrconformantstructmemorysize':( 'int', None, 'stdcall', 'rpcrt4.NdrConformantStructMemorySize', ( ('int', None), ('int', None)) ),
    'rpcrt4.ndrconformantstructunmarshall':( 'int', None, 'stdcall', 'rpcrt4.NdrConformantStructUnmarshall', (('void *', 'ptr'), ('void *', 'ptr'), ('int', None), ('void *', 'ptr')) ),
    'rpcrt4.ndrconformantvaryingarraybuffersize':( 'int', None, 'stdcall', 'rpcrt4.NdrConformantVaryingArrayBufferSize', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrconformantvaryingarrayfree':( 'int', None, 'stdcall', 'rpcrt4.NdrConformantVaryingArrayFree', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrconformantvaryingarraymarshall':( 'int', None, 'stdcall', 'rpcrt4.NdrConformantVaryingArrayMarshall', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrconformantvaryingarraymemorysize':( 'int', None, 'stdcall', 'rpcrt4.NdrConformantVaryingArrayMemorySize', (('int', None), ('int', None)) ),
    'rpcrt4.ndrconformantvaryingarrayunmarshall':( 'int', None, 'stdcall', 'rpcrt4.NdrConformantVaryingArrayUnmarshall', (('int', None), ('void *', 'ptr'), ('int', None), ('int', None)) ),
    'rpcrt4.ndrconformantvaryingstructbuffersize':( 'int', None, 'stdcall', 'rpcrt4.NdrConformantVaryingStructBufferSize', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrconformantvaryingstructfree':( 'int', None, 'stdcall', 'rpcrt4.NdrConformantVaryingStructFree', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrconformantvaryingstructmarshall':( 'int', None, 'stdcall', 'rpcrt4.NdrConformantVaryingStructMarshall', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrconformantvaryingstructmemorysize':( 'int', None, 'stdcall', 'rpcrt4.NdrConformantVaryingStructMemorySize', (('void *', 'ptr'), ('int', None)) ),
    'rpcrt4.ndrconformantvaryingstructunmarshall':( 'int', None, 'stdcall', 'rpcrt4.NdrConformantVaryingStructUnmarshall', (('void *', 'ptr'), ('int', None), ('void *', 'obj'), ('int', None)) ),
    'rpcrt4.ndrcontexthandleinitialize':( 'int', None, 'stdcall', 'rpcrt4.NdrContextHandleInitialize', (('int', None), ('int', None)) ),
    'rpcrt4.ndrcontexthandlesize':( 'int', None, 'stdcall', 'rpcrt4.NdrContextHandleSize', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrconvert':( 'int', None, 'stdcall', 'rpcrt4.NdrConvert', (('void *', 'ptr'), ('int', None)) ),
    'rpcrt4.ndrconvert2':( 'int', None, 'stdcall', 'rpcrt4.NdrConvert2', (('void *', 'ptr'), ('int', None), ('int', None)) ),
    'rpcrt4.ndrcorrelationfree':( 'int', None, 'stdcall', 'rpcrt4.NdrCorrelationFree', (('void *', 'ptr'),) ),
    'rpcrt4.ndrcorrelationinitialize':( 'int', None, 'stdcall', 'rpcrt4.NdrCorrelationInitialize', (('int', None), ('void *', 'ptr'), ('int', None), ('int', None)) ),
    'rpcrt4.ndrcorrelationpass':( 'int', None, 'stdcall', 'rpcrt4.NdrCorrelationPass', (('int', None),) ),
    'rpcrt4.ndrcreateserverinterfacefromstub':( 'int', None, 'stdcall', 'rpcrt4.NdrCreateServerInterfaceFromStub', (('int', None), ('int', None)) ),
    'rpcrt4.ndrdcomasyncclientcall':( 'int', None, 'cdecl', 'rpcrt4.NdrDcomAsyncClientCall', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrdcomasyncstubcall':( 'int', None, 'stdcall', 'rpcrt4.NdrDcomAsyncStubCall', (('int', None), ('int', None), ('int', None), ('void *', 'ptr')) ),
    'rpcrt4.ndrdllcanunloadnow':( 'int', None, 'stdcall', 'rpcrt4.NdrDllCanUnloadNow', (('int', None),) ),
    'rpcrt4.ndrdllgetclassobject':( 'int', None, 'stdcall', 'rpcrt4.NdrDllGetClassObject', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrdllregisterproxy':( 'int', None, 'stdcall', 'rpcrt4.NdrDllRegisterProxy', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrdllunregisterproxy':( 'int', None, 'stdcall', 'rpcrt4.NdrDllUnregisterProxy', (('int', None), ('int', None), ('void *', 'obj')) ),
    'rpcrt4.ndrencapsulatedunionbuffersize':( 'int', None, 'stdcall', 'rpcrt4.NdrEncapsulatedUnionBufferSize', (('void *', 'ptr'), ('int', None), ('int', None)) ),
    'rpcrt4.ndrencapsulatedunionfree':( 'int', None, 'stdcall', 'rpcrt4.NdrEncapsulatedUnionFree', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrencapsulatedunionmarshall':( 'int', None, 'stdcall', 'rpcrt4.NdrEncapsulatedUnionMarshall', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrencapsulatedunionmemorysize':( 'int', None, 'stdcall', 'rpcrt4.NdrEncapsulatedUnionMemorySize', (('int', None), ('int', None)) ),
    'rpcrt4.ndrencapsulatedunionunmarshall':( 'int', None, 'stdcall', 'rpcrt4.NdrEncapsulatedUnionUnmarshall', (('void *', 'ptr'), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrfixedarraybuffersize':( 'int', None, 'stdcall', 'rpcrt4.NdrFixedArrayBufferSize', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrfixedarrayfree':( 'int', None, 'stdcall', 'rpcrt4.NdrFixedArrayFree', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrfixedarraymarshall':( 'int', None, 'stdcall', 'rpcrt4.NdrFixedArrayMarshall', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrfixedarraymemorysize':( 'int', None, 'stdcall', 'rpcrt4.NdrFixedArrayMemorySize', (('int', None), ('int', None)) ),
    'rpcrt4.ndrfixedarrayunmarshall':( 'int', None, 'stdcall', 'rpcrt4.NdrFixedArrayUnmarshall', (('void *', 'ptr'), ('int', None), ('int', None), ('void *', 'ptr')) ),
    'rpcrt4.ndrfreebuffer':( 'int', None, 'stdcall', 'rpcrt4.NdrFreeBuffer', (('void *', 'ptr'),) ),
    'rpcrt4.ndrfullpointerfree':( 'int', None, 'stdcall', 'rpcrt4.NdrFullPointerFree', (('int', None), ('int', None)) ),
    'rpcrt4.ndrfullpointerinsertrefid':( 'int', None, 'stdcall', 'rpcrt4.NdrFullPointerInsertRefId', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrfullpointerquerypointer':( 'int', None, 'stdcall', 'rpcrt4.NdrFullPointerQueryPointer', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrfullpointerqueryrefid':( 'int', None, 'stdcall', 'rpcrt4.NdrFullPointerQueryRefId', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrfullpointerxlatfree':( 'int', None, 'stdcall', 'rpcrt4.NdrFullPointerXlatFree', (('void *', 'ptr'),) ),
    'rpcrt4.ndrfullpointerxlatinit':( 'int', None, 'stdcall', 'rpcrt4.NdrFullPointerXlatInit', (('DWORD', None), ('int', None)) ),
    'rpcrt4.ndrgetbuffer':( 'int', None, 'stdcall', 'rpcrt4.NdrGetBuffer', (('void *', 'ptr'), ('int', None), ('int', None)) ),
    'rpcrt4.ndrgetdcomprotocolversion':( 'int', None, 'stdcall', 'rpcrt4.NdrGetDcomProtocolVersion', (('int', None), ('int', None)) ),
    'rpcrt4.ndrgetsimpletypebufferalignment':( 'int', None, 'stdcall', 'rpcrt4.NdrGetSimpleTypeBufferAlignment', (('int', None),) ),
    'rpcrt4.ndrgetsimpletypebuffersize':( 'int', None, 'stdcall', 'rpcrt4.NdrGetSimpleTypeBufferSize', (('int', None),) ),
    'rpcrt4.ndrgetsimpletypememorysize':( 'int', None, 'stdcall', 'rpcrt4.NdrGetSimpleTypeMemorySize', (('int', None),) ),
    'rpcrt4.ndrgettypeflags':( 'int', None, 'stdcall', 'rpcrt4.NdrGetTypeFlags', (('int', None),) ),
    'rpcrt4.ndrgetusermarshalinfo':( 'int', None, 'stdcall', 'rpcrt4.NdrGetUserMarshalInfo', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrinterfacepointerbuffersize':( 'int', None, 'stdcall', 'rpcrt4.NdrInterfacePointerBufferSize', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrinterfacepointerfree':( 'int', None, 'stdcall', 'rpcrt4.NdrInterfacePointerFree', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrinterfacepointermarshall':( 'int', None, 'stdcall', 'rpcrt4.NdrInterfacePointerMarshall', (('int', None), ('void *', 'obj'), ('int', None)) ),
    'rpcrt4.ndrinterfacepointermemorysize':( 'int', None, 'stdcall', 'rpcrt4.NdrInterfacePointerMemorySize', (('int', None), ('int', None)) ),
    'rpcrt4.ndrinterfacepointerunmarshall':( 'int', None, 'stdcall', 'rpcrt4.NdrInterfacePointerUnmarshall', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrmapcommandfaultstatus':( 'int', None, 'stdcall', 'rpcrt4.NdrMapCommAndFaultStatus', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrmesprocencodedecode':( 'int', None, 'cdecl', 'rpcrt4.NdrMesProcEncodeDecode', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrmesprocencodedecode2':( 'int', None, 'cdecl', 'rpcrt4.NdrMesProcEncodeDecode2', (('void *', 'ptr'), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrmessimpletypealignsize':( 'int', None, 'stdcall', 'rpcrt4.NdrMesSimpleTypeAlignSize', (('int', None),) ),
    'rpcrt4.ndrmessimpletypedecode':( 'int', None, 'stdcall', 'rpcrt4.NdrMesSimpleTypeDecode', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrmessimpletypeencode':( 'int', None, 'stdcall', 'rpcrt4.NdrMesSimpleTypeEncode', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrmestypealignsize':( 'int', None, 'stdcall', 'rpcrt4.NdrMesTypeAlignSize', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrmestypealignsize2':( 'int', None, 'stdcall', 'rpcrt4.NdrMesTypeAlignSize2', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrmestypedecode':( 'int', None, 'stdcall', 'rpcrt4.NdrMesTypeDecode', (('void *', 'ptr'), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrmestypedecode2':( 'int', None, 'stdcall', 'rpcrt4.NdrMesTypeDecode2', (('void *', 'ptr'), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrmestypeencode':( 'int', None, 'stdcall', 'rpcrt4.NdrMesTypeEncode', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrmestypeencode2':( 'int', None, 'stdcall', 'rpcrt4.NdrMesTypeEncode2', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrmestypefree2':( 'int', None, 'stdcall', 'rpcrt4.NdrMesTypeFree2', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrnonconformantstringbuffersize':( 'int', None, 'stdcall', 'rpcrt4.NdrNonConformantStringBufferSize', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrnonconformantstringmarshall':( 'int', None, 'stdcall', 'rpcrt4.NdrNonConformantStringMarshall', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrnonconformantstringmemorysize':( 'int', None, 'stdcall', 'rpcrt4.NdrNonConformantStringMemorySize', (('int', None), ('int', None)) ),
    'rpcrt4.ndrnonconformantstringunmarshall':( 'int', None, 'stdcall', 'rpcrt4.NdrNonConformantStringUnmarshall', (('void *', 'ptr'), ('void *', 'ptr'), ('int', None), ('void *', 'ptr')) ),
    'rpcrt4.ndrnonencapsulatedunionbuffersize':( 'int', None, 'stdcall', 'rpcrt4.NdrNonEncapsulatedUnionBufferSize', (('void *', 'ptr'), ('int', None), ('int', None)) ),
    'rpcrt4.ndrnonencapsulatedunionfree':( 'int', None, 'stdcall', 'rpcrt4.NdrNonEncapsulatedUnionFree', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrnonencapsulatedunionmarshall':( 'int', None, 'stdcall', 'rpcrt4.NdrNonEncapsulatedUnionMarshall', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrnonencapsulatedunionmemorysize':( 'int', None, 'stdcall', 'rpcrt4.NdrNonEncapsulatedUnionMemorySize', (('int', None), ('int', None)) ),
    'rpcrt4.ndrnonencapsulatedunionunmarshall':( 'int', None, 'stdcall', 'rpcrt4.NdrNonEncapsulatedUnionUnmarshall', (('void *', 'ptr'), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrnsgetbuffer':( 'int', None, 'stdcall', 'rpcrt4.NdrNsGetBuffer', (('void *', 'ptr'), ('int', None), ('int', None)) ),
    'rpcrt4.ndrnssendreceive':( 'int', None, 'stdcall', 'rpcrt4.NdrNsSendReceive', (('void *', 'ptr'), ('int', None), ('int', None)) ),
    'rpcrt4.ndroleallocate':( 'int', None, 'stdcall', 'rpcrt4.NdrOleAllocate', (('int', None),) ),
    'rpcrt4.ndrolefree':( 'int', None, 'stdcall', 'rpcrt4.NdrOleFree', (('int', None),) ),
    'rpcrt4.ndroutinit':( 'int', None, 'stdcall', 'rpcrt4.NdrOutInit', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrpartialignoreclientbuffersize':( 'int', None, 'stdcall', 'rpcrt4.NdrPartialIgnoreClientBufferSize', (('int', None), ('int', None)) ),
    'rpcrt4.ndrpartialignoreclientmarshall':( 'int', None, 'stdcall', 'rpcrt4.NdrPartialIgnoreClientMarshall', (('int', None), ('int', None)) ),
    'rpcrt4.ndrpartialignoreserverinitialize':( 'int', None, 'stdcall', 'rpcrt4.NdrPartialIgnoreServerInitialize', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrpartialignoreserverunmarshall':( 'int', None, 'stdcall', 'rpcrt4.NdrPartialIgnoreServerUnmarshall', (('int', None), ('int', None)) ),
    'rpcrt4.ndrpointerbuffersize':( 'int', None, 'stdcall', 'rpcrt4.NdrPointerBufferSize', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrpointerfree':( 'int', None, 'stdcall', 'rpcrt4.NdrPointerFree', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrpointermarshall':( 'int', None, 'stdcall', 'rpcrt4.NdrPointerMarshall', (('int', None), ('void *', 'obj'), ('int', None)) ),
    'rpcrt4.ndrpointermemorysize':( 'int', None, 'stdcall', 'rpcrt4.NdrPointerMemorySize', (('int', None), ('int', None)) ),
    'rpcrt4.ndrpointerunmarshall':( 'int', None, 'stdcall', 'rpcrt4.NdrPointerUnmarshall', (('void *', 'ptr'), ('void *', 'ptr'), ('void *', 'ptr'), ('int', None)) ),
    'rpcrt4.ndrproxyerrorhandler':( 'int', None, 'stdcall', 'rpcrt4.NdrProxyErrorHandler', (('int', None),) ),
    'rpcrt4.ndrproxyfreebuffer':( 'int', None, 'stdcall', 'rpcrt4.NdrProxyFreeBuffer', (('int', None), ('void *', 'ptr')) ),
    'rpcrt4.ndrproxygetbuffer':( 'int', None, 'stdcall', 'rpcrt4.NdrProxyGetBuffer', (('int', None), ('void *', 'ptr')) ),
    'rpcrt4.ndrproxyinitialize':( 'int', None, 'stdcall', 'rpcrt4.NdrProxyInitialize', (('int', None), ('void *', 'ptr'), ('void *', 'ptr'), ('int', None), ('int', None)) ),
    'rpcrt4.ndrproxysendreceive':( 'int', None, 'stdcall', 'rpcrt4.NdrProxySendReceive', (('int', None), ('void *', 'ptr')) ),
    'rpcrt4.ndrrangeunmarshall':( 'int', None, 'stdcall', 'rpcrt4.NdrRangeUnmarshall', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrrpcsmclientallocate':( 'int', None, 'stdcall', 'rpcrt4.NdrRpcSmClientAllocate', (('int', None),) ),
    'rpcrt4.ndrrpcsmclientfree':( 'int', None, 'stdcall', 'rpcrt4.NdrRpcSmClientFree', (('int', None),) ),
    'rpcrt4.ndrrpcsmsetclienttoosf':( 'int', None, 'stdcall', 'rpcrt4.NdrRpcSmSetClientToOsf', (('void *', 'ptr'),) ),
    'rpcrt4.ndrrpcssdefaultallocate':( 'int', None, 'stdcall', 'rpcrt4.NdrRpcSsDefaultAllocate', (('DWORD', None),) ),
    'rpcrt4.ndrrpcssdefaultfree':( 'int', None, 'stdcall', 'rpcrt4.NdrRpcSsDefaultFree', (('void *', 'ptr'),) ),
    'rpcrt4.ndrrpcssdisableallocate':( 'int', None, 'stdcall', 'rpcrt4.NdrRpcSsDisableAllocate', (('void *', 'ptr'),) ),
    'rpcrt4.ndrrpcssenableallocate':( 'int', None, 'stdcall', 'rpcrt4.NdrRpcSsEnableAllocate', (('int', None),) ),
    'rpcrt4.ndrsendreceive':( 'int', None, 'stdcall', 'rpcrt4.NdrSendReceive', (('void *', 'ptr'), ('int', None)) ),
    'rpcrt4.ndrservercall':( 'int', None, 'stdcall', 'rpcrt4.NdrServerCall', (('int', None),) ),
    'rpcrt4.ndrservercall2':( 'int', None, 'stdcall', 'rpcrt4.NdrServerCall2', (('int', None),) ),
    'rpcrt4.ndrservercontextmarshall':( 'int', None, 'stdcall', 'rpcrt4.NdrServerContextMarshall', (('int', None), ('void *', 'obj'), ('int', None)) ),
    'rpcrt4.ndrservercontextnewmarshall':( 'int', None, 'stdcall', 'rpcrt4.NdrServerContextNewMarshall', (('int', None), ('void *', 'obj'), ('int', None), ('int', None)) ),
    'rpcrt4.ndrservercontextnewunmarshall':( 'int', None, 'stdcall', 'rpcrt4.NdrServerContextNewUnmarshall', (('void *', 'ptr'), ('void *', 'ptr')) ),
    'rpcrt4.ndrservercontextunmarshall':( 'int', None, 'stdcall', 'rpcrt4.NdrServerContextUnmarshall', (('int', None),) ),
    'rpcrt4.ndrserverinitialize':( 'int', None, 'stdcall', 'rpcrt4.NdrServerInitialize', (('int', None), ('void *', 'ptr'), ('int', None)) ),
    'rpcrt4.ndrserverinitializemarshall':( 'int', None, 'stdcall', 'rpcrt4.NdrServerInitializeMarshall', (('int', None), ('int', None)) ),
    'rpcrt4.ndrserverinitializenew':( 'int', None, 'stdcall', 'rpcrt4.NdrServerInitializeNew', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrserverinitializepartial':( 'int', None, 'stdcall', 'rpcrt4.NdrServerInitializePartial', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrserverinitializeunmarshall':( 'int', None, 'stdcall', 'rpcrt4.NdrServerInitializeUnmarshall', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrservermarshall':( 'int', None, 'stdcall', 'rpcrt4.NdrServerMarshall', (('int', None), ('int', None), ('void *', 'ptr'), ('int', None)) ),
    'rpcrt4.ndrserverunmarshall':( 'int', None, 'stdcall', 'rpcrt4.NdrServerUnmarshall', (('int', None), ('int', None), ('void *', 'ptr'), ('int', None), ('int', None), ('void *', 'ptr')) ),
    'rpcrt4.ndrsimplestructbuffersize':( 'int', None, 'stdcall', 'rpcrt4.NdrSimpleStructBufferSize', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrsimplestructfree':( 'int', None, 'stdcall', 'rpcrt4.NdrSimpleStructFree', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrsimplestructmarshall':( 'int', None, 'stdcall', 'rpcrt4.NdrSimpleStructMarshall', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrsimplestructmemorysize':( 'int', None, 'stdcall', 'rpcrt4.NdrSimpleStructMemorySize', (('int', None), ('void *', 'obj')) ),
    'rpcrt4.ndrsimplestructunmarshall':( 'int', None, 'stdcall', 'rpcrt4.NdrSimpleStructUnmarshall', (('void *', 'ptr'), ('void *', 'ptr'), ('int', None), ('void *', 'ptr')) ),
    'rpcrt4.ndrsimpletypemarshall':( 'int', None, 'stdcall', 'rpcrt4.NdrSimpleTypeMarshall', (('void *', 'ptr'), ('int', None), ('int', None)) ),
    'rpcrt4.ndrsimpletypeunmarshall':( 'int', None, 'stdcall', 'rpcrt4.NdrSimpleTypeUnmarshall', (('void *', 'ptr'), ('void *', 'ptr'), ('int', None)) ),
    'rpcrt4.ndrstubcall':( 'int', None, 'stdcall', 'rpcrt4.NdrStubCall', (('int', None), ('int', None), ('int', None), ('void *', 'ptr')) ),
    'rpcrt4.ndrstubcall2':( 'int', None, 'stdcall', 'rpcrt4.NdrStubCall2', (('int', None), ('int', None), ('int', None), ('void *', 'ptr')) ),
    'rpcrt4.ndrstubforwardingfunction':( 'int', None, 'stdcall', 'rpcrt4.NdrStubForwardingFunction', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrstubgetbuffer':( 'int', None, 'stdcall', 'rpcrt4.NdrStubGetBuffer', (('int', None), ('int', None), ('void *', 'ptr')) ),
    'rpcrt4.ndrstubinitialize':( 'int', None, 'stdcall', 'rpcrt4.NdrStubInitialize', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrstubinitializemarshall':( 'int', None, 'stdcall', 'rpcrt4.NdrStubInitializeMarshall', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrtypeflags':( 'int', None, 'cdecl', 'rpcrt4.NdrTypeFlags', () ),
    'rpcrt4.ndrtypefree':( 'int', None, 'stdcall', 'rpcrt4.NdrTypeFree', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrtypemarshall':( 'int', None, 'stdcall', 'rpcrt4.NdrTypeMarshall', (('void *', 'ptr'), ('int', None), ('int', None)) ),
    'rpcrt4.ndrtypesize':( 'int', None, 'stdcall', 'rpcrt4.NdrTypeSize', (('void *', 'ptr'), ('int', None), ('int', None)) ),
    'rpcrt4.ndrtypeunmarshall':( 'int', None, 'stdcall', 'rpcrt4.NdrTypeUnmarshall', (('int', None), ('void *', 'ptr'), ('int', None), ('int', None)) ),
    'rpcrt4.ndrunmarshallbasetypeinline':( 'int', None, 'stdcall', 'rpcrt4.NdrUnmarshallBasetypeInline', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrusermarshalbuffersize':( 'int', None, 'stdcall', 'rpcrt4.NdrUserMarshalBufferSize', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrusermarshalfree':( 'int', None, 'stdcall', 'rpcrt4.NdrUserMarshalFree', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrusermarshalmarshall':( 'int', None, 'stdcall', 'rpcrt4.NdrUserMarshalMarshall', (('int', None), ('void *', 'obj'), ('int', None)) ),
    'rpcrt4.ndrusermarshalmemorysize':( 'int', None, 'stdcall', 'rpcrt4.NdrUserMarshalMemorySize', (('int', None), ('int', None)) ),
    'rpcrt4.ndrusermarshalsimpletypeconvert':( 'int', None, 'stdcall', 'rpcrt4.NdrUserMarshalSimpleTypeConvert', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrusermarshalunmarshall':( 'int', None, 'stdcall', 'rpcrt4.NdrUserMarshalUnmarshall', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrvaryingarraybuffersize':( 'int', None, 'stdcall', 'rpcrt4.NdrVaryingArrayBufferSize', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrvaryingarrayfree':( 'int', None, 'stdcall', 'rpcrt4.NdrVaryingArrayFree', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrvaryingarraymarshall':( 'int', None, 'stdcall', 'rpcrt4.NdrVaryingArrayMarshall', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrvaryingarraymemorysize':( 'int', None, 'stdcall', 'rpcrt4.NdrVaryingArrayMemorySize', (('int', None), ('int', None)) ),
    'rpcrt4.ndrvaryingarrayunmarshall':( 'int', None, 'stdcall', 'rpcrt4.NdrVaryingArrayUnmarshall', (('void *', 'ptr'), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrxmitorrepasbuffersize':( 'int', None, 'stdcall', 'rpcrt4.NdrXmitOrRepAsBufferSize', (('void *', 'ptr'), ('int', None), ('int', None)) ),
    'rpcrt4.ndrxmitorrepasfree':( 'int', None, 'stdcall', 'rpcrt4.NdrXmitOrRepAsFree', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrxmitorrepasmarshall':( 'int', None, 'stdcall', 'rpcrt4.NdrXmitOrRepAsMarshall', (('void *', 'ptr'), ('int', None), ('int', None)) ),
    'rpcrt4.ndrxmitorrepasmemorysize':( 'int', None, 'stdcall', 'rpcrt4.NdrXmitOrRepAsMemorySize', (('int', None), ('int', None)) ),
    'rpcrt4.ndrxmitorrepasunmarshall':( 'int', None, 'stdcall', 'rpcrt4.NdrXmitOrRepAsUnmarshall', (('void *', 'ptr'), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrpcreateproxy':( 'int', None, 'stdcall', 'rpcrt4.NdrpCreateProxy', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrpcreatestub':( 'int', None, 'stdcall', 'rpcrt4.NdrpCreateStub', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrpgetprocformatstring':( 'int', None, 'stdcall', 'rpcrt4.NdrpGetProcFormatString', (('void *', 'ptr'), ('int', None), ('int', None), ('int', None), ('int', None), ('void *', 'ptr')) ),
    'rpcrt4.ndrpgettypeformatstring':( 'int', None, 'stdcall', 'rpcrt4.NdrpGetTypeFormatString', ( ('void *', 'ptr'), ('int', None)) ),
    'rpcrt4.ndrpgettypegencookie':( 'int', None, 'stdcall', 'rpcrt4.NdrpGetTypeGenCookie', (('int', None),) ),
    'rpcrt4.ndrpmemoryincrement':( 'int', None, 'stdcall', 'rpcrt4.NdrpMemoryIncrement', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.ndrpreleasetypeformatstring':( 'int', None, 'stdcall', 'rpcrt4.NdrpReleaseTypeFormatString', (('void *', 'ptr'),) ),
    'rpcrt4.ndrpreleasetypegencookie':( 'int', None, 'stdcall', 'rpcrt4.NdrpReleaseTypeGenCookie', () ),
    'rpcrt4.ndrpsetrpcssdefaults':( 'int', None, 'stdcall', 'rpcrt4.NdrpSetRpcSsDefaults', (('int', None), ('int', None)) ),
    'rpcrt4.ndrpvarvtoftypedesc':( 'int', None, 'stdcall', 'rpcrt4.NdrpVarVtOfTypeDesc', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.rpcasyncabortcall':( 'int', None, 'stdcall', 'rpcrt4.RpcAsyncAbortCall', (('void *', 'ptr'), ('int', None)) ),
    'rpcrt4.rpcasynccancelcall':( 'int', None, 'stdcall', 'rpcrt4.RpcAsyncCancelCall', (('int', None), ('int', None)) ),
    'rpcrt4.rpcasynccompletecall':( 'int', None, 'stdcall', 'rpcrt4.RpcAsyncCompleteCall', (('void *', 'ptr'), ('void *', 'ptr')) ),
    'rpcrt4.rpcasyncgetcallstatus':( 'int', None, 'stdcall', 'rpcrt4.RpcAsyncGetCallStatus', (('int', None),) ),
    'rpcrt4.rpcasyncinitializehandle':( 'int', None, 'stdcall', 'rpcrt4.RpcAsyncInitializeHandle', (('int', None), ('int', None)) ),
    'rpcrt4.rpcasyncregisterinfo':( 'int', None, 'stdcall', 'rpcrt4.RpcAsyncRegisterInfo', (('int', None),) ),
    'rpcrt4.rpcbindingcopy':( 'int', None, 'stdcall', 'rpcrt4.RpcBindingCopy', ( ('int', None),) ),
    'rpcrt4.rpcbindingfree':( 'int', None, 'stdcall', 'rpcrt4.RpcBindingFree', (('void *', 'ptr'),) ),
    'rpcrt4.rpcbindingfromstringbindinga':( 'int', None, 'stdcall', 'rpcrt4.RpcBindingFromStringBindingA', (('int', None), ('void *', 'ptr')) ),
    'rpcrt4.rpcbindingfromstringbindingw':( 'int', None, 'stdcall', 'rpcrt4.RpcBindingFromStringBindingW', (('void *', 'ptr'), ('int', None)) ),
    'rpcrt4.rpcbindinginqauthclienta':( 'int', None, 'stdcall', 'rpcrt4.RpcBindingInqAuthClientA', ( ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.rpcbindinginqauthclientexa':( 'int', None, 'stdcall', 'rpcrt4.RpcBindingInqAuthClientExA', ( ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.rpcbindinginqauthclientexw':( 'int', None, 'stdcall', 'rpcrt4.RpcBindingInqAuthClientExW', ( ('int', None), ('void *', 'ptr'), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.rpcbindinginqauthclientw':( 'int', None, 'stdcall', 'rpcrt4.RpcBindingInqAuthClientW', ( ('int', None), ('void *', 'ptr'), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.rpcbindinginqauthinfoa':( 'int', None, 'stdcall', 'rpcrt4.RpcBindingInqAuthInfoA', ( ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.rpcbindinginqauthinfoexa':( 'int', None, 'stdcall', 'rpcrt4.RpcBindingInqAuthInfoExA', ( ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.rpcbindinginqauthinfoexw':( 'int', None, 'stdcall', 'rpcrt4.RpcBindingInqAuthInfoExW', ( ('void *', 'ptr'), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.rpcbindinginqauthinfow':( 'int', None, 'stdcall', 'rpcrt4.RpcBindingInqAuthInfoW', ( ('void *', 'ptr'), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.rpcbindinginqobject':( 'int', None, 'stdcall', 'rpcrt4.RpcBindingInqObject', (('int', None), ('void *', 'ptr')) ),
    'rpcrt4.rpcbindinginqoption':( 'int', None, 'stdcall', 'rpcrt4.RpcBindingInqOption', ( ('int', None), ('int', None)) ),
    'rpcrt4.rpcbindingreset':( 'int', None, 'stdcall', 'rpcrt4.RpcBindingReset', () ),
    'rpcrt4.rpcbindingserverfromclient':( 'int', None, 'stdcall', 'rpcrt4.RpcBindingServerFromClient', ( ('void *', 'ptr'),) ),
    'rpcrt4.rpcbindingsetauthinfoa':( 'int', None, 'stdcall', 'rpcrt4.RpcBindingSetAuthInfoA', ( ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.rpcbindingsetauthinfoexa':( 'int', None, 'stdcall', 'rpcrt4.RpcBindingSetAuthInfoExA', ( ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.rpcbindingsetauthinfoexw':( 'int', None, 'stdcall', 'rpcrt4.RpcBindingSetAuthInfoExW', ( ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.rpcbindingsetauthinfow':( 'int', None, 'stdcall', 'rpcrt4.RpcBindingSetAuthInfoW', ( ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.rpcbindingsetobject':( 'int', None, 'stdcall', 'rpcrt4.RpcBindingSetObject', ( ('void *', 'obj'),) ),
    'rpcrt4.rpcbindingsetoption':( 'int', None, 'stdcall', 'rpcrt4.RpcBindingSetOption', ( ('int', None), ('int', None)) ),
    'rpcrt4.rpcbindingtostringbindinga':( 'int', None, 'stdcall', 'rpcrt4.RpcBindingToStringBindingA', ( ('void *', 'ptr'),) ),
    'rpcrt4.rpcbindingtostringbindingw':( 'int', None, 'stdcall', 'rpcrt4.RpcBindingToStringBindingW', ( ('void *', 'ptr'),) ),
    'rpcrt4.rpcbindingvectorfree':( 'int', None, 'stdcall', 'rpcrt4.RpcBindingVectorFree', (('void *', 'ptr'),) ),
    'rpcrt4.rpccancelthread':( 'int', None, 'stdcall', 'rpcrt4.RpcCancelThread', (('int', None),) ),
    'rpcrt4.rpccancelthreadex':( 'int', None, 'stdcall', 'rpcrt4.RpcCancelThreadEx', (('int', None), ('int', None)) ),
    'rpcrt4.rpccertgenerateprincipalnamea':( 'int', None, 'stdcall', 'rpcrt4.RpcCertGeneratePrincipalNameA', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.rpccertgenerateprincipalnamew':( 'int', None, 'stdcall', 'rpcrt4.RpcCertGeneratePrincipalNameW', (('int', None), ('int', None), ('void *', 'ptr')) ),
    'rpcrt4.rpcepregistera':( 'int', None, 'stdcall', 'rpcrt4.RpcEpRegisterA', (('int', None), ('int', None), ('void *', 'ptr'), ('int', None)) ),
    'rpcrt4.rpcepregisternoreplacea':( 'int', None, 'stdcall', 'rpcrt4.RpcEpRegisterNoReplaceA', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.rpcepregisternoreplacew':( 'int', None, 'stdcall', 'rpcrt4.RpcEpRegisterNoReplaceW', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.rpcepregisterw':( 'int', None, 'stdcall', 'rpcrt4.RpcEpRegisterW', (('int', None), ('int', None), ('void *', 'ptr'), ('int', None)) ),
    'rpcrt4.rpcepresolvebinding':( 'int', None, 'stdcall', 'rpcrt4.RpcEpResolveBinding', ( ('int', None),) ),
    'rpcrt4.rpcepunregister':( 'int', None, 'stdcall', 'rpcrt4.RpcEpUnregister', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.rpcerroraddrecord':( 'int', None, 'stdcall', 'rpcrt4.RpcErrorAddRecord', (('int', None),) ),
    'rpcrt4.rpcerrorclearinformation':( 'int', None, 'cdecl', 'rpcrt4.RpcErrorClearInformation', () ),
    'rpcrt4.rpcerrorendenumeration':( 'int', None, 'stdcall', 'rpcrt4.RpcErrorEndEnumeration', (('int', None),) ),
    'rpcrt4.rpcerrorgetnextrecord':( 'int', None, 'stdcall', 'rpcrt4.RpcErrorGetNextRecord', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.rpcerrorgetnumberofrecords':( 'int', None, 'stdcall', 'rpcrt4.RpcErrorGetNumberOfRecords', (('int', None), ('int', None)) ),
    'rpcrt4.rpcerrorloaderrorinfo':( 'int', None, 'stdcall', 'rpcrt4.RpcErrorLoadErrorInfo', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.rpcerrorresetenumeration':( 'int', None, 'stdcall', 'rpcrt4.RpcErrorResetEnumeration', (('int', None),) ),
    'rpcrt4.rpcerrorsaveerrorinfo':( 'int', None, 'stdcall', 'rpcrt4.RpcErrorSaveErrorInfo', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.rpcerrorstartenumeration':( 'int', None, 'stdcall', 'rpcrt4.RpcErrorStartEnumeration', (('int', None),) ),
    'rpcrt4.rpcfreeauthorizationcontext':( 'int', None, 'stdcall', 'rpcrt4.RpcFreeAuthorizationContext', (('int', None),) ),
    'rpcrt4.rpcgetauthorizationcontextforclient':( 'int', None, 'stdcall', 'rpcrt4.RpcGetAuthorizationContextForClient', ( ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.rpcifidvectorfree':( 'int', None, 'stdcall', 'rpcrt4.RpcIfIdVectorFree', (('int', None),) ),
    'rpcrt4.rpcifinqid':( 'int', None, 'stdcall', 'rpcrt4.RpcIfInqId', (('int', None), ('void *', 'ptr')) ),
    'rpcrt4.rpcimpersonateclient':( 'int', None, 'stdcall', 'rpcrt4.RpcImpersonateClient', () ),
    'rpcrt4.rpcmgmtenableidlecleanup':( 'int', None, 'cdecl', 'rpcrt4.RpcMgmtEnableIdleCleanup', () ),
    'rpcrt4.rpcmgmtepeltinqbegin':( 'int', None, 'stdcall', 'rpcrt4.RpcMgmtEpEltInqBegin', ( ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.rpcmgmtepeltinqdone':( 'int', None, 'stdcall', 'rpcrt4.RpcMgmtEpEltInqDone', (('int', None),) ),
    'rpcrt4.rpcmgmtepeltinqnexta':( 'int', None, 'stdcall', 'rpcrt4.RpcMgmtEpEltInqNextA', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.rpcmgmtepeltinqnextw':( 'int', None, 'stdcall', 'rpcrt4.RpcMgmtEpEltInqNextW', (('int', None), ('int', None), ('int', None), ('void *', 'ptr'), ('int', None)) ),
    'rpcrt4.rpcmgmtepunregister':( 'int', None, 'stdcall', 'rpcrt4.RpcMgmtEpUnregister', ( ('void *', 'obj'), ('void *', 'ptr'), ('void *', 'obj'), ('int', None)) ),
    'rpcrt4.rpcmgmtinqcomtimeout':( 'int', None, 'stdcall', 'rpcrt4.RpcMgmtInqComTimeout', ( ('void *', 'ptr'),) ),
    'rpcrt4.rpcmgmtinqdefaultprotectlevel':( 'int', None, 'stdcall', 'rpcrt4.RpcMgmtInqDefaultProtectLevel', ( ('int', None), ('int', None)) ),
    'rpcrt4.rpcmgmtinqifids':( 'int', None, 'stdcall', 'rpcrt4.RpcMgmtInqIfIds', (('int', None), ('int', None)) ),
    'rpcrt4.rpcmgmtinqserverprincnamea':( 'int', None, 'stdcall', 'rpcrt4.RpcMgmtInqServerPrincNameA', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.rpcmgmtinqserverprincnamew':( 'int', None, 'stdcall', 'rpcrt4.RpcMgmtInqServerPrincNameW', (('int', None), ('int', None), ('void *', 'ptr')) ),
    'rpcrt4.rpcmgmtinqstats':( 'int', None, 'stdcall', 'rpcrt4.RpcMgmtInqStats', (('int', None), ('int', None)) ),
    'rpcrt4.rpcmgmtisserverlistening':( 'int', None, 'stdcall', 'rpcrt4.RpcMgmtIsServerListening', (('int', None),) ),
    'rpcrt4.rpcmgmtsetauthorizationfn':( 'int', None, 'stdcall', 'rpcrt4.RpcMgmtSetAuthorizationFn', (('int', None),) ),
    'rpcrt4.rpcmgmtsetcanceltimeout':( 'int', None, 'stdcall', 'rpcrt4.RpcMgmtSetCancelTimeout', (('int', None),) ),
    'rpcrt4.rpcmgmtsetcomtimeout':( 'int', None, 'stdcall', 'rpcrt4.RpcMgmtSetComTimeout', ( ('int', None),) ),
    'rpcrt4.rpcmgmtsetserverstacksize':( 'int', None, 'stdcall', 'rpcrt4.RpcMgmtSetServerStackSize', (('int', None),) ),
    'rpcrt4.rpcmgmtstatsvectorfree':( 'int', None, 'stdcall', 'rpcrt4.RpcMgmtStatsVectorFree', (('void *', 'ptr'),) ),
    'rpcrt4.rpcmgmtstopserverlistening':( 'int', None, 'stdcall', 'rpcrt4.RpcMgmtStopServerListening', (('int', None),) ),
    'rpcrt4.rpcmgmtwaitserverlisten':( 'int', None, 'cdecl', 'rpcrt4.RpcMgmtWaitServerListen', () ),
    'rpcrt4.rpcnetworkinqprotseqsa':( 'int', None, 'stdcall', 'rpcrt4.RpcNetworkInqProtseqsA', ( ('void *', 'ptr'),) ),
    'rpcrt4.rpcnetworkinqprotseqsw':( 'int', None, 'stdcall', 'rpcrt4.RpcNetworkInqProtseqsW', (('int', None),) ),
    'rpcrt4.rpcnetworkisprotseqvalida':( 'int', None, 'stdcall', 'rpcrt4.RpcNetworkIsProtseqValidA', (('int', None),) ),
    'rpcrt4.rpcnetworkisprotseqvalidw':( 'int', None, 'stdcall', 'rpcrt4.RpcNetworkIsProtseqValidW', (('void *', 'ptr'),) ),
    'rpcrt4.rpcnsbindinginqentrynamea':( 'int', None, 'stdcall', 'rpcrt4.RpcNsBindingInqEntryNameA', ( ('int', None), ('int', None)) ),
    'rpcrt4.rpcnsbindinginqentrynamew':( 'int', None, 'stdcall', 'rpcrt4.RpcNsBindingInqEntryNameW', ( ('int', None), ('void *', 'ptr')) ),
    'rpcrt4.rpcobjectinqtype':( 'int', None, 'stdcall', 'rpcrt4.RpcObjectInqType', ( ('int', None),) ),
    'rpcrt4.rpcobjectsetinqfn':( 'int', None, 'stdcall', 'rpcrt4.RpcObjectSetInqFn', (('int', None),) ),
    'rpcrt4.rpcobjectsettype':( 'int', None, 'stdcall', 'rpcrt4.RpcObjectSetType', ( ('void *', 'obj'),) ),
    'rpcrt4.rpcprotseqvectorfreea':( 'int', None, 'stdcall', 'rpcrt4.RpcProtseqVectorFreeA', (('int', None),) ),
    'rpcrt4.rpcprotseqvectorfreew':( 'int', None, 'stdcall', 'rpcrt4.RpcProtseqVectorFreeW', (('void *', 'ptr'),) ),
    'rpcrt4.rpcraiseexception':( 'int', None, 'stdcall', 'rpcrt4.RpcRaiseException', (('int', None),) ),
    'rpcrt4.rpcreverttoself':( 'int', None, 'cdecl', 'rpcrt4.RpcRevertToSelf', () ),
    'rpcrt4.rpcreverttoselfex':( 'int', None, 'stdcall', 'rpcrt4.RpcRevertToSelfEx', () ),
    'rpcrt4.rpcserverinqbindings':( 'int', None, 'stdcall', 'rpcrt4.RpcServerInqBindings', (('int', None),) ),
    'rpcrt4.rpcserverinqcallattributesa':( 'int', None, 'stdcall', 'rpcrt4.RpcServerInqCallAttributesA', ( ('int', None),) ),
    'rpcrt4.rpcserverinqcallattributesw':( 'int', None, 'stdcall', 'rpcrt4.RpcServerInqCallAttributesW', ( ('void *', 'ptr'),) ),
    'rpcrt4.rpcserverinqdefaultprincnamea':( 'int', None, 'stdcall', 'rpcrt4.RpcServerInqDefaultPrincNameA', ( ('int', None), ('int', None)) ),
    'rpcrt4.rpcserverinqdefaultprincnamew':( 'int', None, 'stdcall', 'rpcrt4.RpcServerInqDefaultPrincNameW', ( ('int', None), ('void *', 'ptr')) ),
    'rpcrt4.rpcserverinqif':( 'int', None, 'stdcall', 'rpcrt4.RpcServerInqIf', (('int', None), ('void *', 'obj'), ('int', None)) ),
    'rpcrt4.rpcserverlisten':( 'int', None, 'stdcall', 'rpcrt4.RpcServerListen', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.rpcserverregisterauthinfoa':( 'int', None, 'stdcall', 'rpcrt4.RpcServerRegisterAuthInfoA', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.rpcserverregisterauthinfow':( 'int', None, 'stdcall', 'rpcrt4.RpcServerRegisterAuthInfoW', (('void *', 'ptr'), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.rpcserverregisterif':( 'int', None, 'stdcall', 'rpcrt4.RpcServerRegisterIf', (('int', None), ('void *', 'obj'), ('int', None)) ),
    'rpcrt4.rpcserverregisterif2':( 'int', None, 'stdcall', 'rpcrt4.RpcServerRegisterIf2', (('int', None), ('void *', 'obj'), ('int', None), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.rpcserverregisterifex':( 'int', None, 'stdcall', 'rpcrt4.RpcServerRegisterIfEx', (('int', None), ('void *', 'obj'), ('int', None), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.rpcservertestcancel':( 'int', None, 'stdcall', 'rpcrt4.RpcServerTestCancel', () ),
    'rpcrt4.rpcserverunregisterif':( 'int', None, 'stdcall', 'rpcrt4.RpcServerUnregisterIf', (('int', None), ('void *', 'obj'), ('int', None)) ),
    'rpcrt4.rpcserverunregisterifex':( 'int', None, 'stdcall', 'rpcrt4.RpcServerUnregisterIfEx', (('int', None), ('void *', 'obj'), ('int', None)) ),
    'rpcrt4.rpcserveruseallprotseqs':( 'int', None, 'stdcall', 'rpcrt4.RpcServerUseAllProtseqs', ( ('int', None), ('void *', 'ptr')) ),
    'rpcrt4.rpcserveruseallprotseqsex':( 'int', None, 'stdcall', 'rpcrt4.RpcServerUseAllProtseqsEx', ( ('int', None), ('void *', 'ptr'), ('void *', 'ptr')) ),
    'rpcrt4.rpcserveruseallprotseqsif':( 'int', None, 'stdcall', 'rpcrt4.RpcServerUseAllProtseqsIf', (('int', None), ('int', None), ('void *', 'ptr')) ),
    'rpcrt4.rpcserveruseallprotseqsifex':( 'int', None, 'stdcall', 'rpcrt4.RpcServerUseAllProtseqsIfEx', (('int', None), ('int', None), ('void *', 'ptr'), ('void *', 'ptr')) ),
    'rpcrt4.rpcserveruseprotseqa':( 'int', None, 'stdcall', 'rpcrt4.RpcServerUseProtseqA', (('int', None), ('int', None), ('void *', 'ptr')) ),
    'rpcrt4.rpcserveruseprotseqepa':( 'int', None, 'stdcall', 'rpcrt4.RpcServerUseProtseqEpA', (('int', None), ('int', None), ('int', None), ('void *', 'ptr')) ),
    'rpcrt4.rpcserveruseprotseqepexa':( 'int', None, 'stdcall', 'rpcrt4.RpcServerUseProtseqEpExA', (('int', None), ('int', None), ('int', None), ('void *', 'ptr'), ('int', None)) ),
    'rpcrt4.rpcserveruseprotseqepexw':( 'int', None, 'stdcall', 'rpcrt4.RpcServerUseProtseqEpExW', (('void *', 'ptr'), ('int', None), ('void *', 'ptr'), ('void *', 'ptr'), ('void *', 'ptr')) ),
    'rpcrt4.rpcserveruseprotseqepw':( 'int', None, 'stdcall', 'rpcrt4.RpcServerUseProtseqEpW', (('void *', 'ptr'), ('int', None), ('void *', 'ptr'), ('void *', 'ptr')) ),
    'rpcrt4.rpcserveruseprotseqexa':( 'int', None, 'stdcall', 'rpcrt4.RpcServerUseProtseqExA', (('int', None), ('int', None), ('void *', 'ptr'), ('int', None)) ),
    'rpcrt4.rpcserveruseprotseqexw':( 'int', None, 'stdcall', 'rpcrt4.RpcServerUseProtseqExW', (('void *', 'ptr'), ('int', None), ('void *', 'ptr'), ('void *', 'ptr')) ),
    'rpcrt4.rpcserveruseprotseqifa':( 'int', None, 'stdcall', 'rpcrt4.RpcServerUseProtseqIfA', (('int', None), ('int', None), ('int', None), ('void *', 'ptr')) ),
    'rpcrt4.rpcserveruseprotseqifexa':( 'int', None, 'stdcall', 'rpcrt4.RpcServerUseProtseqIfExA', (('int', None), ('int', None), ('int', None), ('void *', 'ptr'), ('void *', 'ptr')) ),
    'rpcrt4.rpcserveruseprotseqifexw':( 'int', None, 'stdcall', 'rpcrt4.RpcServerUseProtseqIfExW', (('void *', 'ptr'), ('int', None), ('int', None), ('void *', 'ptr'), ('void *', 'ptr')) ),
    'rpcrt4.rpcserveruseprotseqifw':( 'int', None, 'stdcall', 'rpcrt4.RpcServerUseProtseqIfW', (('void *', 'ptr'), ('int', None), ('int', None), ('void *', 'ptr')) ),
    'rpcrt4.rpcserveruseprotseqw':( 'int', None, 'stdcall', 'rpcrt4.RpcServerUseProtseqW', (('void *', 'ptr'), ('int', None), ('void *', 'ptr')) ),
    'rpcrt4.rpcserveryield':( 'int', None, 'cdecl', 'rpcrt4.RpcServerYield', () ),
    'rpcrt4.rpcsmallocate':( 'int', None, 'stdcall', 'rpcrt4.RpcSmAllocate', (('int', None), ('int', None)) ),
    'rpcrt4.rpcsmclientfree':( 'int', None, 'stdcall', 'rpcrt4.RpcSmClientFree', (('int', None),) ),
    'rpcrt4.rpcsmdestroyclientcontext':( 'int', None, 'stdcall', 'rpcrt4.RpcSmDestroyClientContext', (('int', None),) ),
    'rpcrt4.rpcsmdisableallocate':( 'int', None, 'cdecl', 'rpcrt4.RpcSmDisableAllocate', () ),
    'rpcrt4.rpcsmenableallocate':( 'int', None, 'cdecl', 'rpcrt4.RpcSmEnableAllocate', () ),
    'rpcrt4.rpcsmfree':( 'int', None, 'stdcall', 'rpcrt4.RpcSmFree', (('int', None),) ),
    'rpcrt4.rpcsmgetthreadhandle':( 'int', None, 'stdcall', 'rpcrt4.RpcSmGetThreadHandle', (('int', None),) ),
    'rpcrt4.rpcsmsetclientallocfree':( 'int', None, 'stdcall', 'rpcrt4.RpcSmSetClientAllocFree', (('int', None), ('int', None)) ),
    'rpcrt4.rpcsmsetthreadhandle':( 'int', None, 'stdcall', 'rpcrt4.RpcSmSetThreadHandle', (('int', None),) ),
    'rpcrt4.rpcsmswapclientallocfree':( 'int', None, 'stdcall', 'rpcrt4.RpcSmSwapClientAllocFree', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.rpcssallocate':( 'int', None, 'stdcall', 'rpcrt4.RpcSsAllocate', (('int', None),) ),
    'rpcrt4.rpcsscontextlockexclusive':( 'int', None, 'stdcall', 'rpcrt4.RpcSsContextLockExclusive', ( ('int', None),) ),
    'rpcrt4.rpcsscontextlockshared':( 'int', None, 'stdcall', 'rpcrt4.RpcSsContextLockShared', ( ('int', None),) ),
    'rpcrt4.rpcssdestroyclientcontext':( 'int', None, 'stdcall', 'rpcrt4.RpcSsDestroyClientContext', (('void *', 'ptr'),) ),
    'rpcrt4.rpcssdisableallocate':( 'int', None, 'cdecl', 'rpcrt4.RpcSsDisableAllocate', () ),
    'rpcrt4.rpcssenableallocate':( 'int', None, 'cdecl', 'rpcrt4.RpcSsEnableAllocate', () ),
    'rpcrt4.rpcssfree':( 'int', None, 'stdcall', 'rpcrt4.RpcSsFree', (('int', None),) ),
    'rpcrt4.rpcssgetcontextbinding':( 'int', None, 'stdcall', 'rpcrt4.RpcSsGetContextBinding', (('int', None), ('int', None)) ),
    'rpcrt4.rpcssgetthreadhandle':( 'int', None, 'cdecl', 'rpcrt4.RpcSsGetThreadHandle', () ),
    'rpcrt4.rpcsssetclientallocfree':( 'int', None, 'stdcall', 'rpcrt4.RpcSsSetClientAllocFree', (('int', None), ('int', None)) ),
    'rpcrt4.rpcsssetthreadhandle':( 'int', None, 'stdcall', 'rpcrt4.RpcSsSetThreadHandle', (('int', None),) ),
    'rpcrt4.rpcssswapclientallocfree':( 'int', None, 'stdcall', 'rpcrt4.RpcSsSwapClientAllocFree', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.rpcstringbindingcomposea':( 'int', None, 'stdcall', 'rpcrt4.RpcStringBindingComposeA', (('int', None), ('int', None), ('int', None), ('int', None), ('int', None), ('void *', 'ptr')) ),
    'rpcrt4.rpcstringbindingcomposew':( 'int', None, 'stdcall', 'rpcrt4.RpcStringBindingComposeW', (('void *', 'ptr'), ('int', None), ('void *', 'ptr'), ('void *', 'ptr'), ('int', None), ('void *', 'ptr')) ),
    'rpcrt4.rpcstringbindingparsea':( 'int', None, 'stdcall', 'rpcrt4.RpcStringBindingParseA', (('int', None), ('int', None), ('void *', 'ptr'), ('void *', 'ptr'), ('void *', 'ptr'), ('int', None)) ),
    'rpcrt4.rpcstringbindingparsew':( 'int', None, 'stdcall', 'rpcrt4.RpcStringBindingParseW', (('int', None), ('void *', 'ptr'), ('void *', 'ptr'), ('void *', 'ptr'), ('void *', 'ptr'), ('void *', 'ptr')) ),
    'rpcrt4.rpcstringfreea':( 'int', None, 'stdcall', 'rpcrt4.RpcStringFreeA', (('void *', 'ptr'),) ),
    'rpcrt4.rpcstringfreew':( 'int', None, 'stdcall', 'rpcrt4.RpcStringFreeW', (('void *', 'ptr'),) ),
    'rpcrt4.rpctestcancel':( 'int', None, 'cdecl', 'rpcrt4.RpcTestCancel', () ),
    'rpcrt4.rpcuserfree':( 'int', None, 'stdcall', 'rpcrt4.RpcUserFree', (('int', None), ('int', None)) ),
    'rpcrt4.simpletypealignment':( 'int', None, 'cdecl', 'rpcrt4.SimpleTypeAlignment', () ),
    'rpcrt4.simpletypebuffersize':( 'int', None, 'msfastcall', 'rpcrt4.SimpleTypeBufferSize', (('int', None), ('int', None)) ),
    'rpcrt4.simpletypememorysize':( 'int', None, 'msfastcall', 'rpcrt4.SimpleTypeMemorySize', (('int', None), ('int', None)) ),
    'rpcrt4.towerconstruct':( 'int', None, 'stdcall', 'rpcrt4.TowerConstruct', ( ('void *', 'ptr'), ('void *', 'ptr'), ('void *', 'ptr'), ('int', None), ('int', None), ('void *', 'ptr')) ),
    'rpcrt4.towerexplode':( 'int', None, 'stdcall', 'rpcrt4.TowerExplode', (('int', None), ('int', None), ('int', None), ('void *', 'ptr'), ('void *', 'ptr'), ('void *', 'ptr')) ),
    'rpcrt4.uuidcompare':( 'int', None, 'stdcall', 'rpcrt4.UuidCompare', ( ('void *', 'obj'), ('int', None)) ),
    'rpcrt4.uuidcreate':( 'int', None, 'stdcall', 'rpcrt4.UuidCreate', (('void *', 'ptr'),) ),
    'rpcrt4.uuidcreatenil':( 'int', None, 'stdcall', 'rpcrt4.UuidCreateNil', () ),
    'rpcrt4.uuidcreatesequential':( 'int', None, 'stdcall', 'rpcrt4.UuidCreateSequential', (('int', None),) ),
    'rpcrt4.uuidequal':( 'int', None, 'stdcall', 'rpcrt4.UuidEqual', ( ('void *', 'obj'), ('int', None)) ),
    'rpcrt4.uuidfromstringa':( 'int', None, 'stdcall', 'rpcrt4.UuidFromStringA', (('int', None), ('void *', 'obj')) ),
    'rpcrt4.uuidfromstringw':( 'int', None, 'stdcall', 'rpcrt4.UuidFromStringW', (('int', None), ('void *', 'obj')) ),
    'rpcrt4.uuidhash':( 'int', None, 'stdcall', 'rpcrt4.UuidHash', ( ('int', None),) ),
    'rpcrt4.uuidisnil':( 'int', None, 'stdcall', 'rpcrt4.UuidIsNil', ( ('void *', 'ptr'),) ),
    'rpcrt4.uuidtostringa':( 'int', None, 'stdcall', 'rpcrt4.UuidToStringA', ( ('int', None),) ),
    'rpcrt4.uuidtostringw':( 'int', None, 'stdcall', 'rpcrt4.UuidToStringW', ( ('void *', 'ptr'),) ),
    'rpcrt4.char_array_from_ndr':( 'int', None, 'stdcall', 'rpcrt4.char_array_from_ndr', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.char_from_ndr':( 'int', None, 'stdcall', 'rpcrt4.char_from_ndr', (('int', None), ('int', None)) ),
    'rpcrt4.data_from_ndr':( 'int', None, 'stdcall', 'rpcrt4.data_from_ndr', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.data_into_ndr':( 'int', None, 'stdcall', 'rpcrt4.data_into_ndr', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.data_size_ndr':( 'int', None, 'stdcall', 'rpcrt4.data_size_ndr', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.double_array_from_ndr':( 'int', None, 'stdcall', 'rpcrt4.double_array_from_ndr', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.double_from_ndr':( 'int', None, 'stdcall', 'rpcrt4.double_from_ndr', (('int', None), ('int', None)) ),
    'rpcrt4.enum_from_ndr':( 'int', None, 'stdcall', 'rpcrt4.enum_from_ndr', (('int', None), ('int', None)) ),
    'rpcrt4.float_array_from_ndr':( 'int', None, 'stdcall', 'rpcrt4.float_array_from_ndr', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.float_from_ndr':( 'int', None, 'stdcall', 'rpcrt4.float_from_ndr', (('int', None), ('int', None)) ),
    'rpcrt4.long_array_from_ndr':( 'int', None, 'stdcall', 'rpcrt4.long_array_from_ndr', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.long_from_ndr':( 'int', None, 'stdcall', 'rpcrt4.long_from_ndr', (('int', None), ('int', None)) ),
    'rpcrt4.long_from_ndr_temp':( 'int', None, 'stdcall', 'rpcrt4.long_from_ndr_temp', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.pfnfreeroutines':( 'int', None, 'cdecl', 'rpcrt4.pfnFreeRoutines', () ),
    'rpcrt4.pfnmarshallroutines':( 'int', None, 'cdecl', 'rpcrt4.pfnMarshallRoutines', () ),
    'rpcrt4.pfnsizeroutines':( 'int', None, 'cdecl', 'rpcrt4.pfnSizeRoutines', () ),
    'rpcrt4.pfnunmarshallroutines':( 'int', None, 'cdecl', 'rpcrt4.pfnUnmarshallRoutines', () ),
    'rpcrt4.short_array_from_ndr':( 'int', None, 'stdcall', 'rpcrt4.short_array_from_ndr', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.short_from_ndr':( 'int', None, 'stdcall', 'rpcrt4.short_from_ndr', (('int', None), ('int', None)) ),
    'rpcrt4.short_from_ndr_temp':( 'int', None, 'stdcall', 'rpcrt4.short_from_ndr_temp', (('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.tree_into_ndr':( 'int', None, 'stdcall', 'rpcrt4.tree_into_ndr', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.tree_peek_ndr':( 'int', None, 'stdcall', 'rpcrt4.tree_peek_ndr', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    'rpcrt4.tree_size_ndr':( 'int', None, 'stdcall', 'rpcrt4.tree_size_ndr', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    # FIXME unresolved I_RpcAbortAsyncCall -> RPCRT4.I_RpcAsyncAbortCall
    # FIXME unresolved I_RpcBindingInqDynamicEndpoint -> RPCRT4.I_RpcBindingInqDynamicEndpointW
    # FIXME unresolved I_RpcNsBindingSetEntryName -> RPCRT4.I_RpcNsBindingSetEntryNameW
    # FIXME unresolved I_RpcSetAsyncHandle -> RPCRT4.I_RpcAsyncSetHandle
    # FIXME unresolved RpcAbortAsyncCall -> RPCRT4.RpcAsyncAbortCall
    # FIXME unresolved RpcCancelAsyncCall -> RPCRT4.RpcAsyncCancelCall
    # FIXME unresolved RpcCompleteAsyncCall -> RPCRT4.RpcAsyncCompleteCall
    # FIXME unresolved RpcGetAsyncCallStatus -> RPCRT4.RpcAsyncGetCallStatus
    # FIXME unresolved RpcInitializeAsyncHandle -> RPCRT4.RpcAsyncInitializeHandle
    # FIXME unresolved RpcRegisterAsyncInfo -> RPCRT4.RpcAsyncRegisterInfo
    # FIXME unresolved RpcSsDontSerializeContext -> RPCRT4.I_RpcSsDontSerializeContext
    }



'''
The cobra "auth" package allows modular authentication
for cobra shared objects
'''

class CobraAuthenticator:

    def __init__(self):
        pass

    def authCobraUser(self, authinfo):
        '''
        Authenticate a user and return their user name.
        Arguments:
            authinfo - a dictionary which can be authmod dependant
                       ( but probably just has user & passwd )
        '''
        # This is *totally* example code... you probably want something
        # from a submodule like cobra.auth.shadowfile
        if authinfo.get('user') != 'invisigoth':
            return None
        if authinfo.get('passwd') != 'secret':
            return None
        return 'invisigoth'

    def checkUserAccess(self, authuser, objname ):
        '''
        Enforce access control on a per shared object basis.
        '''
        return True




"""
Home of the i386 module's register specs/code.
"""
import envi.registers as e_reg

## Definitions for some of the i386 MSRs from intel...
MSR_DEBUGCTL             = 0x01d9 # Intel p4 and forward, debug behavior control
MSR_DEBUGCTL_LBR         = 0x0001 # last branch recording (in msr's)
MSR_DEBUGCTL_BTF         = 0x0002 # single-step on branches (break on branch)
MSR_DEBUGCTL_TR          = 0x0004 # enable sending "branch trace messages" !!
MSR_DEBUGCTL_BTS         = 0x0008 # enable logging BTMs to circular buffer
MSR_DEBUGCTL_BTINT       = 0x0010 # Branch-trace-interrupt (gen interrupt on BTS full)
MSR_DEBUGCTL_BTS_OFF_OS  = 0x0020 # disable ring0 branch trace store
MSR_DEBUGCTL_BTS_OFF_USR = 0x0040 # disable non-ring0 branch trace store

MSR_SYSENTER_EIP         = 0x0176 # Where is EIP at sysenter?

IA32_DS_AREA_MSR         = 0x0600 # pointer to the configured debug storage area

i386regs = [
    ("eax",32),("ecx",32),("edx",32),("ebx",32),("esp",32),("ebp",32),("esi",32),("edi",32),
    # SIMD registers
    ("xmm0",128),("xmm1",128),("xmm2",128),("xmm3",128),("xmm4",128),("xmm5",128),("xmm6",128),("xmm7",128),
    # Debug registers
    ("debug0",32),("debug1",32),("debug2",32),("debug3",32),("debug4",32),("debug5",32),("debug6",32),("debug7",32),
    # Control registers
    ("ctrl0",32),("ctrl1",32),("ctrl2",32),("ctrl3",32),("ctrl4",32),("ctrl5",32),("ctrl6",32),("ctrl7",32),
    # Test registers
    ("test0", 32),("test1", 32),("test2", 32),("test3", 32),("test4", 32),("test5", 32),("test6", 32),("test7", 32),
    # Segment registers
    ("es", 16),("cs",16),("ss",16),("ds",16),("fs",16),("gs",16),
    # FPU Registers
    ("st0", 80),("st1", 80),("st2", 80),("st3", 80),("st4", 80),("st5", 80),("st6", 80),("st7", 80),
    # Leftovers ;)
    ("eflags", 32), ("eip", 32), ("fpsr", 16), ("fpcr", 16),
    # TODO there's a bunch of floating point stuff that we basically just ignore
]

def getRegOffset(regs, regname):
    # NOTE: dynamically calculate this on import so we are less
    # likely to fuck it up...
    for i,(name,width) in enumerate(regs):
        if name == regname:
            return i
    raise Exception("getRegOffset doesn't know about: %s" % regname)

# dynamically create REG_EAX and the like in our module
l = locals()
e_reg.addLocalEnums(l, i386regs)

i386meta = [
    ("mm0", REG_ST0, 0, 64),
    ("mm1", REG_ST1, 0, 64),
    ("mm2", REG_ST2, 0, 64),
    ("mm3", REG_ST3, 0, 64),
    ("mm4", REG_ST4, 0, 64),
    ("mm5", REG_ST5, 0, 64),
    ("mm6", REG_ST6, 0, 64),
    ("mm7", REG_ST7, 0, 64),
    ("ax", REG_EAX, 0, 16),
    ("cx", REG_ECX, 0, 16),
    ("dx", REG_EDX, 0, 16),
    ("bx", REG_EBX, 0, 16),
    ("sp", REG_ESP, 0, 16),
    ("bp", REG_EBP, 0, 16),
    ("si", REG_ESI, 0, 16),
    ("di", REG_EDI, 0, 16),

    ("al", REG_EAX, 0, 8),
    ("cl", REG_ECX, 0, 8),
    ("dl", REG_EDX, 0, 8),
    ("bl", REG_EBX, 0, 8),

    ("ah", REG_EAX, 8, 8),
    ("ch", REG_ECX, 8, 8),
    ("dh", REG_EDX, 8, 8),
    ("bh", REG_EBX, 8, 8),
]

statmetas = [
        ('CF', REG_EFLAGS, 0, 1, 'Carry Flag'),
        ('PF', REG_EFLAGS, 2, 1, 'Parity Flag'),
        ('AF', REG_EFLAGS, 4, 1, 'Adjust Flag'),
        ('ZF', REG_EFLAGS, 6, 1, 'Zero Flag'),
        ('SF', REG_EFLAGS, 7, 1, 'Sign Flag'),
        ('TF', REG_EFLAGS, 8, 1, 'Trap Flag'),
        ('IF', REG_EFLAGS, 9, 1, 'Interrupt Enable Flag'),
        ('DF', REG_EFLAGS, 10, 1, 'Direction Flag'),
        ('OF', REG_EFLAGS, 11, 1, 'Overflow Flag'),
        ('IOPL', REG_EFLAGS, 12, 2, 'I/O Privilege Level'),
        ('NT', REG_EFLAGS, 14, 1, 'Nested Task'),
        ('RF', REG_EFLAGS, 16, 1, 'Resume Flag'),
        ('VM', REG_EFLAGS, 17, 1, 'Virtual-8086 Mode'),
        ('AC', REG_EFLAGS, 18, 1, 'Alignment Check'),
        ('VIF', REG_EFLAGS, 19, 1, 'Virtual Interrupt Flag'),
        ('VIP', REG_EFLAGS, 20, 1, 'Virtual Interrupt Pending'),
        ('ID', REG_EFLAGS, 21, 1, 'ID Flag'),
        ]

def getEflagsFields(regval):
    ret = []
    for name,_,shift,bits,desc in statmetas:
        ret.append( (name, regval >> shift & 1) )
    return ret

e_reg.addLocalStatusMetas(l, i386meta, statmetas, 'EFLAGS')
e_reg.addLocalMetas(l, i386meta)

class i386RegisterContext(e_reg.RegisterContext):
    def __init__(self):
        e_reg.RegisterContext.__init__(self)
        self.loadRegDef(i386regs)
        self.loadRegMetas(i386meta, statmetas=statmetas)
        self.setRegisterIndexes(REG_EIP, REG_ESP, srindex=REG_EFLAGS)



"""

Cobra's distributed code module capable of allowing
serialization of code from one system to another.

Particularly useful for clustering and workunit stuff.

"""
import os
import sys
import logging
import importlib

import cobra

from importlib.abc import Loader, MetaPathFinder
from importlib.util import spec_from_loader, module_from_spec

logger = logging.getLogger(__name__)


class DcodeServer:
    '''
    Server-side Dcode class.  Allows the sharing of Server-side python modules to the remote client
    '''

    def getPythonModule(self, fullname, path=None):

        # Some serialization causes this to be a tuple
        # and find_module is SUPER SERIOUS about it being
        # a *list*... ;)
        if path is not None:
            path = list(path)

        fullname = fullname.split(".")[-1]

        try:
            mspec = importlib.util.find_spec(fullname)
            filename = mspec.origin
        except ImportError:
            return None

        if not os.path.exists(filename):
            return None

        path = os.path.dirname(filename)
        with open(filename, 'r') as f:
            fbytes = f.read()
        return (fbytes, filename, path)


class DcodeLoader(Loader):
    """
    This object gets returned by the DcodeFinder (client-side)
    """
    def __init__(self, fbytes, filename, path):
        object.__init__(self)
        self.fbytes = fbytes
        self.filename = filename
        self.path = path

    def get_source(self, name):
        return self.fbytes

    def load_module(self, fullname):
        mod = sys.modules.get(fullname)
        if mod is None:
            spec = spec_from_loader(fullname, loader=None)
            mod = module_from_spec(spec)
            sys.modules[fullname] = mod
            exec(self.fbytes, mod.__dict__)
            mod.__file__ = self.filename
            mod.__loader__ = self
            if self.path is not None:
                mod.__path__ = [self.path]

        return mod


    def create_module(self, spec):
        return None

    def exec_module(self, mod):
        exec(self.fbytes, mod.__dict__)
        mod.__file__ = self.filename
        mod.__loader__ = self
        if self.path is not None:
            mod.__path__ = [self.path]
        


class DcodeFinder(object):
    """
    This object goes into the client side import path_hooks
    to allow cobra:// uri's to be added to the import path.

    For reference, this is a MetaPathFinder
    """
    def __init__(self, proxy):
        self.proxy = proxy
        self.counter = 0

    def find_module(self, fullname, path=None):

        # Check for local modules first...
        localname = fullname.split('.')[-1]
        name, ext = os.path.splitext(localname)

        try:
            mspec = importlib.util.find_spec(fullname)
        except ValueError:

            logger.info('Dcode Searching: %s (%s)', name, path)
            pymod = self.proxy.getPythonModule(fullname, path)
            if pymod:
                logger.info('Dcode Loaded: %s', fullname)
                return DcodeLoader(*pymod)

    def find_spec(self, fullname, path, target=None):
        if self.counter:
            raise AttributeError("find_spec called by itself....  stop recursion")
        self.counter += 1

        # Check for local modules first...
        localname = fullname.split('.')[-1]
        name, ext = os.path.splitext(localname)

        try:
            mspec = importlib.util.find_spec(fullname, path)

        except AttributeError as e:
            logger.info('Dcode Searching: %s (%s)', name, path)
            pymod = self.proxy.getPythonModule(fullname, path)
            if pymod:
                logger.info('Dcode Loaded: %s', fullname)
                # need to return a spec
                return spec_from_loader(fullname, DcodeLoader(*pymod))

        finally:
            self.counter -= 1

# Client-side helper functions
def addDcodeProxy(proxy):
    finder = DcodeFinder(proxy)
    sys.meta_path.append(finder)


def addDcodeUri(uri):
    proxy = cobra.CobraProxy(uri, timeout=120, retrymax=3)
    addDcodeProxy(proxy)


def addDcodeServer(server, port=cobra.COBRA_PORT, ssl=False, msgpack=True):
    scheme = "cobra"
    if ssl:
        scheme = "cobrassl"

    uri = "%s://%s:%d/DcodeServer" % (scheme, server, port)
    if msgpack:
        uri += "?msgpack=1"

    addDcodeUri(uri)


# Server-side helper function
def enableDcodeServer(daemon=None):
    server = DcodeServer()
    if daemon:
        daemon.shareObject(server, 'DcodeServer')
        return
    cobra.shareObject(server, 'DcodeServer')



def main():
    '''
    Launch a Dcode server including any paths you list on the command line
    '''
    import argparse
    parser = argparse.ArgumentParser(prog='cobraDcode', usage='%(prog)s [options] [additional_path1 [additional_path2] [...]]>')
    parser.add_argument('-C', '--clearpath', default=False, help='Clear the PYTHONPATH and only share specified paths')
    parser.add_argument('-P', '--port', dest='port', type=int, default=cobra.COBRA_PORT,
                        help='Listen on what port')
    parser.add_argument('--cacrt', default=None, help='Use TLS encryption: With this Certificate Authority Cert (file)')
    parser.add_argument('--srvcrt', default=None, help='Use TLS encryption: With this Server Cert (file)')
    parser.add_argument('--srvkey', default=None, help='Use TLS encryption: With this Server Key (file)')
    parser.add_argument('path', nargs='*')
    args = parser.parse_args()

    if args.clearpath:
        sys.path = []

    for path in args.path:
        if path not in sys.path:
            logger.info('Adding path "%s" to system path (ie. will serve to Dcode clients)', path)
            sys.path.append(path)

    # first start a cobra server on the specified port
    daemon = cobra.startCobraServer(port=args.port, sslca=args.cacrt, sslcrt=args.srvcrt, sslkey=args.srvkey, msgpack=1)

    # next, enable Dcode
    enableDcodeServer(daemon)

    # wait forever (or until the process is killed)
    daemon.thr.join()




if __name__ == "__main__":
    import envi.common as ecmn
    ecmn.initLogging(logger, logging.DEBUG)
    main()
